import datetime
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple

from pydantic import Field, PositiveFloat

from avtdl.core.actors import Filter, FilterEntity
from avtdl.core.interfaces import Record
from avtdl.core.monitors import PagedFeedMonitor, PagedFeedMonitorConfig, PagedFeedMonitorEntity
from avtdl.core.plugins import Plugins
from avtdl.core.request import HttpClient, RetrySettings
from avtdl.core.runtime import RuntimeContext
from avtdl.plugins.filters.filters import EmptyFilterConfig
from avtdl.plugins.youtube.common import NextPageContext, get_innertube_context, get_session_index, handle_consent, \
    prepare_next_page_request
from avtdl.plugins.youtube.feed_info import AuthorInfo, ContentTypeNotSupportedException, VideoRendererInfo, \
    get_video_renderers, parse_lockup_view, \
    parse_owner_info, \
    parse_video_renderer


@Plugins.register('channel', Plugins.kind.ASSOCIATED_RECORD)
@Plugins.register('filter.channel', Plugins.kind.ASSOCIATED_RECORD)
class YoutubeVideoRecord(VideoRendererInfo, Record):
    """
    Youtube video or livestream listed among others on Youtube page

    Produced by parsing a channels main page, videos and streams tab,
    as well as playlists, and, with login cookies, subscriptions feed.
    """

    video_id: str
    """short string identifying video on Youtube. Part of video url"""
    url: str
    """link to the video, uses `https://www.youtube.com/watch?v=<video_id>` format"""
    title: str
    """title of the video at the time of parsing"""
    summary: Optional[str] = Field(repr=False)
    """snippet of the video description. Not always available"""
    scheduled: Optional[datetime.datetime] = None
    """scheduled date for upcoming stream or premiere"""
    author: Optional[str]
    """channel name"""
    avatar_url: Optional[str] = None
    """link to the avatar of the channel. Not always available"""
    thumbnail_url: Optional[str] = None
    """link to the video thumbnail"""
    channel_link: Optional[str] = None
    """link to the channel uploading the video"""
    channel_id: Optional[str] = None
    """channel ID in old format (such as `UCK0V3b23uJyU4N8eR_BR0QA`)"""
    published_text: Optional[str]
    """localized text saying how long ago the video was uploaded"""
    length: Optional[str]
    """text showing the video duration (hh:mm:ss)"""

    is_upcoming: bool
    """indicates that video is an upcoming livestream or premiere"""
    is_live: bool
    """indicates that the video is a livestream or premiere that is currently live"""
    is_member_only: bool
    """indicated that the video is limited to members of the channel. Note that the video status might be changed at any time"""

    def __str__(self):
        last_line = ''
        scheduled = self.scheduled
        if scheduled:
            last_line = '\nscheduled to {}'.format(scheduled.strftime('%Y-%m-%d %H:%M'))
        elif self.is_live:
            last_line = '\n[Live]'
        if self.is_member_only:
            last_line += ' [Member only]'
        template = '{}\n{}\npublished by {}'
        return template.format(self.url, self.title, self.author) + last_line

    def __repr__(self):
        template = '{:<8} [{}] {}'
        return template.format(self.author or 'Unknown author', self.video_id, self.title[:60])

    def get_uid(self) -> str:
        return self.video_id

    def as_embed(self) -> dict:
        embed: Dict[str, Any] = {
            'title': self.title,
            # 'description': ,
            'url': self.url,
            'color': None,
            'author': {'name': self.author, 'url': self.channel_link, 'icon_url': self.avatar_url},
            'image': {'url': self.thumbnail_url},
            'fields': []
        }
        footer = ''
        if self.published_text:
            footer += self.published_text
        if self.length:
            embed['fields'].append({'name': f'[{self.length}]', 'value': '', 'inline': True})
        if self.scheduled is not None:
            scheduled = self.scheduled.strftime('%Y-%m-%d %H:%M')
            embed['fields'].append({'name': 'Scheduled:', 'value': scheduled, 'inline': True})
        if self.is_live:
            embed['fields'].append({'name': '[Live]', 'value': '', 'inline': True})
        if self.is_member_only:
            embed['fields'].append({'name': '[Member only]', 'value': '', 'inline': True})
        embed['footer'] = {'text': footer}
        return embed


@Plugins.register('channel', Plugins.kind.ACTOR_CONFIG)
class VideosMonitorConfig(PagedFeedMonitorConfig):
    pass


@Plugins.register('channel', Plugins.kind.ACTOR_ENTITY)
class VideosMonitorEntity(PagedFeedMonitorEntity):
    update_interval: PositiveFloat = 1800


class FeedPageContext(NextPageContext):
    owner_info: Optional[AuthorInfo]


@Plugins.register('channel', Plugins.kind.ACTOR)
class VideosMonitor(PagedFeedMonitor):
    """
    Youtube channel monitor

    Monitors Youtube url listing videos, such as channels main page,
    videos and streams tab of a channel, as well as playlists, and,
    with login cookies, subscriptions feed or the main page.

    Due to small differences in presentation in aforementioned
    sources, same video might have slightly different appearance when
    parsed from different urls. For example, video parsed from main
    page or subscriptions feed will not have full description text.

    Examples of supported url:

    - `https://www.youtube.com/@ChannelName`
    - `https://www.youtube.com/@ChannelName/videos`
    - `https://www.youtube.com/@ChannelName/streams`
    - `https://www.youtube.com/channel/UCK0V3b23uJyU4N8eR_BR0QA/`
    - `https://www.youtube.com/playlist?list=PLWGY3fcU-ZeQmBfoJ6SmT8v2zV8NEhrB2`
    - `https://www.youtube.com/feed/subscriptions` (providing cookies is necessarily)

    Unlike `rss` monitor, with login cookies it can see videos and streams
    with limited access (such as member-only).

    While monitoring a single channel is less efficient, using this monitor with
    subscriptions feed url on a dedicated account is a recommended way
    to monitor a high amount (hundreds) of channels, as it only requires
    loading a single page to check all of them for updates.

    When main page of a channel (https://www.youtube.com/@ChannelName) is viewed
    in logged in state, it might contain "For you" block, which content might
    vary with subsequent updates. As a result, monitoring this url might occasionally
    produce records with old videos that got showed in this block. If monitoring
    without a cookies file is not an option, use a combination of "Videos" and "Streams"
    tabs instead.
    """

    async def handle_first_page(self, entity: PagedFeedMonitorEntity, client: HttpClient) -> Tuple[Optional[Sequence[Record]], Optional[FeedPageContext]]:
        raw_page_text = await self.request(entity.url, entity, client)
        if raw_page_text is None:
            return None, None
        raw_page_text = await handle_consent(raw_page_text, entity.url, client, self.logger)
        video_renderers, lockup_views, continuation_token, page = get_video_renderers(raw_page_text)
        if not video_renderers and not lockup_views:
            self.logger.warning(f'[{entity.name}] found no videos on first page of {entity.url}')
        owner_info = parse_owner_info(page)
        current_page_records = self._parse_records(owner_info, video_renderers, lockup_views, entity)
        innertube_context = get_innertube_context(raw_page_text)
        session_index = get_session_index(page)
        context = FeedPageContext(innertube_context=innertube_context, session_index=session_index, continuation_token=continuation_token, owner_info=owner_info)
        return current_page_records, context

    async def handle_next_page(self, entity: PagedFeedMonitorEntity, client: HttpClient,
                               context: Optional[FeedPageContext]) -> Tuple[Optional[Sequence[Record]], Optional[FeedPageContext]]:
        if context is None or context.continuation_token is None:
            self.logger.debug(f'[{entity.name}] no continuation for next page, done loading')
            return [], None

        url, headers, post_body = prepare_next_page_request(context.innertube_context, context.continuation_token, cookies=client.cookie_jar)
        raw_page = await client.request_text(url, method='POST', data_json=post_body, headers=headers,
                                        settings=RetrySettings(retry_times=3, retry_delay=5, retry_multiplier=2))
        if raw_page is None:
            self.logger.debug(f'[{entity.name}] failed to load next page, aborting')
            return None, None
        video_renderers, lockup_views, continuation_token, page = get_video_renderers(raw_page, anchor='')

        if not video_renderers and not lockup_views:
            self.logger.debug(f'[{entity.name}] found no videos when parsing continuation of {entity.url}')
        current_page_records = self._parse_records(context.owner_info, video_renderers, lockup_views, entity)

        if continuation_token is not None:
            context.continuation_token = continuation_token
        else:
            context = None
        return current_page_records, context

    def _parse_records(self,
                       owner_info: Optional[AuthorInfo],
                       video_renderers: List[dict],
                       lockup_views: List[dict],
                       entity: PagedFeedMonitorEntity) -> List[YoutubeVideoRecord]:
        records: List[YoutubeVideoRecord] = []
        records.extend(self._parse_entries(video_renderers, parse_video_renderer, owner_info, entity))
        records.extend(self._parse_entries(lockup_views, parse_lockup_view, owner_info, entity))
        records = records[::-1] # records are ordered from old to new on page, reorder in chronological order
        return records

    def _parse_entries(self, items: List[dict],
                       parser: Callable,
                       owner_info: Optional[AuthorInfo],
                       entity: PagedFeedMonitorEntity) -> List[YoutubeVideoRecord]:
        records: List[YoutubeVideoRecord] = []
        for item in items:
            try:
                info = parser(item, owner_info)
                record = YoutubeVideoRecord.model_validate(info.model_dump())
                records.append(record)
            except ContentTypeNotSupportedException as e:
                self.logger.debug(f'skipping unsupported content type: {e}')
                continue
            except Exception as e:
                self.logger.warning(f'[{entity.name}] failed to parse video renderer on "{entity.url}": {type(e)}: {e}')
                self.logger.debug(f'[{entity.name}] raw video renderer:\n{item}')
                continue
        return records

    def record_got_updated(self, record: YoutubeVideoRecord, entity: VideosMonitorEntity) -> bool:
        excluded_fields = {'published_text'}
        return self.db.record_has_changed(record, entity.name, excluded_fields)


@Plugins.register('filter.channel', Plugins.kind.ACTOR_CONFIG)
class ChannelFilterConfig(EmptyFilterConfig):
    pass

@Plugins.register('filter.channel', Plugins.kind.ACTOR_ENTITY)
class ChannelFilterEntity(FilterEntity):
    upcoming: bool = False
    """to pass the filter a record should be either upcoming livestream or scheduled premiere"""
    live: bool = False
    """to pass the filter a record should be an ongoing livestream"""
    member_only: bool = False
    """to pass the filter a record should be marked as member-only"""


@Plugins.register('filter.channel', Plugins.kind.ACTOR)
class ChannelFilter(Filter):
    """
    Pick `YoutubeVideoRecord` with specified properties

    Filter that only lets `YoutubeVideoRecord` through if it has certain properties.
    All records from other sources pass through without filtering.

    If multiple settings are set to `true`, they all should match. Use multiple
    entities if picking records with any of multiple properties is required.
    """

    def __init__(self, config: ChannelFilterConfig, entities: Sequence[ChannelFilterEntity], ctx: RuntimeContext):
        super().__init__(config, entities, ctx)

    def match(self, entity: ChannelFilterEntity, record: YoutubeVideoRecord) -> Optional[YoutubeVideoRecord]:
        if not isinstance(record, YoutubeVideoRecord):
            self.logger.debug(f'[{entity.name}] record is not a YoutubeVideoRecord, letting it through: {record!r}')
            return record
        if entity.upcoming and not record.is_upcoming:
            return None
        if entity.live and not record.is_live:
            return None
        if entity.member_only and not record.is_member_only:
            return None
        return record



import logging
import logging.handlers
from enum import Enum
from pathlib import Path
from typing import Dict, Optional, Union

from avtdl.core.utils import check_dir


class LogLevel(str, Enum):
    debug = 'DEBUG'
    info = 'INFO'
    warning = 'WARNING'
    error = 'ERROR'


class LoggingConfig:
    """
    Static class used to hold references to logger handlers

    Keeping references to handlers allows to call setup methods
    multiple times, enabling restarting the app with new config
    by removing existing handlers before attaching new ones.
    """

    LOG_FORMAT = '%(asctime)s.%(msecs)03d [%(levelname)-7s] [%(name)s] %(message)s'
    ACCESS_LOG_FORMAT = '%(message)s'
    DATE_FORMAT = '%Y/%m/%d %H:%M:%S'

    stream_handler: Optional[logging.StreamHandler] = None
    file_handler: Optional[logging.FileHandler] = None
    access_handler: Optional[logging.FileHandler] = None

    @classmethod
    def setup_console_logger(cls, level):
        root = logging.getLogger()
        if cls.stream_handler is not None:
            root.removeHandler(cls.stream_handler)
        cls.stream_handler = logging.StreamHandler()
        cls.stream_handler.setLevel(level)
        formatter = logging.Formatter(cls.LOG_FORMAT, cls.DATE_FORMAT)
        cls.stream_handler.setFormatter(formatter)
        root.addHandler(cls.stream_handler)
        root.name = 'avtdl'
        root.setLevel(logging.NOTSET)

    @classmethod
    def create_file_handler(cls, path: Path, name: str, max_size, level: LogLevel, log_format: str, date_format: str) -> Optional[logging.FileHandler]:
        check_dir(path, create=True)
        path /= name
        try:
            handler = logging.handlers.RotatingFileHandler(path, maxBytes=max_size, backupCount=10, encoding='utf8')
        except Exception as e:
            logging.error(f'writing log to {path.absolute()} failed: {e}')
            return None
        formatter = logging.Formatter(log_format, date_format)
        handler.setFormatter(formatter)
        handler.setLevel(getattr(logging, level))
        return handler

    @classmethod
    def setup_file_logger(cls, path: Path, name: str, max_size, level: LogLevel):
        handler = cls.create_file_handler(path, name, max_size, level, cls.LOG_FORMAT, cls.DATE_FORMAT)

        logger = logging.getLogger()
        if cls.file_handler is not None:
            logger.removeHandler(cls.file_handler)
        cls.file_handler = handler
        if cls.file_handler is not None:
            logger.addHandler(cls.file_handler)
            logging.info(f'writing verbose log to file {path.absolute()}')

    @classmethod
    def setup_webserver_logger(cls, path: Path, name: str, max_size, level: LogLevel):
        handler = cls.create_file_handler(path, name, max_size, level, cls.ACCESS_LOG_FORMAT, cls.DATE_FORMAT)

        logger = logging.getLogger('aiohttp.access')
        if cls.access_handler is not None:
            logger.removeHandler(cls.access_handler)
        cls.access_handler = handler
        if cls.access_handler is None:
            return
        logger.addHandler(cls.access_handler)
        logger.propagate = False
        logging.info(f'writing access log to file {path.absolute()}')


def setup_console_logger(level):
    """
    Set up stdout logging format

    Called before config was parsed, sets up
    console logger based on command line arguments
    """
    LoggingConfig.setup_console_logger(level)


def setup_file_logger(path: Path, max_size, level: LogLevel):
    """
    Set up logging to file based on configuration parameters
    """
    LoggingConfig.setup_file_logger(path, 'avtdl.log', max_size, level)


def setup_webserver_logger(path: Path, max_size, level: LogLevel):
    """
    Set up access log of the webui webserver
    """
    LoggingConfig.setup_webserver_logger(path, 'access.log', max_size, level)


def set_logger_loglevel(logger_name: str, level: Union[LogLevel, int], propagate=True):
    if isinstance(level, LogLevel):
        # since LogLevel itself should only contain valid log level names
        # it should never raise AttributeError, but if it does let it crash
        # so it can get noticed
        level = getattr(logging, level)
    logger = logging.getLogger(logger_name)
    logger.propagate = propagate
    logger.setLevel(level)


def silence_library_loggers():
    set_logger_loglevel('asyncio', logging.WARNING)
    set_logger_loglevel('charset_normalizer', logging.WARNING)
    set_logger_loglevel('slixmpp', logging.ERROR)


def override_loglevel(loggers: Dict[str, LogLevel]):
    for name, level in loggers.items():
        set_logger_loglevel(name, level, propagate=True)



import base64
import datetime
import hashlib
import http.cookiejar
import http.cookies
import json
import logging
import os
import re
import urllib.parse
from collections import OrderedDict
from contextlib import ContextDecorator
from http import cookiejar
from http.cookiejar import CookieJar
from pathlib import Path
from textwrap import shorten
from time import perf_counter
from typing import Any, Dict, Hashable, Iterable, Iterator, List, Mapping, MutableMapping, Optional, Tuple, \
    Type, TypeVar, Union

import aiohttp
import dateutil.parser
import dateutil.tz
from aiohttp.abc import AbstractCookieJar
from jsonpath import JSONPath
from pydantic import AnyHttpUrl, RootModel, ValidationError

from avtdl.core.interfaces import Record

JSONType = Union[str, int, float, bool, None, Mapping[str, 'JSONType'], List['JSONType']]


def load_cookies(path: Optional[Path], raise_on_error: bool = False) -> Optional[cookiejar.CookieJar]:
    """load cookies from a text file in Netscape format"""
    logger = logging.getLogger('cookies')
    if path is None:
        return None
    cookie_jar = cookiejar.MozillaCookieJar(path)
    try:
        cookie_jar.load(ignore_discard=True, ignore_expires=True)
        logger.info(f"Successfully loaded cookies from {path}")
    except FileNotFoundError:
        logger.exception(f'Failed to load cookies from {path}: file not found')
        if raise_on_error:
            raise
        return None
    except (cookiejar.LoadError, OSError) as e:
        if raise_on_error:
            raise
        logger.exception(f'Failed to load cookies from {path}: {e}')
        return None
    return cookie_jar


class CookieStoreError(Exception):
    """Raised when save_cookies() failed"""


def save_cookies(cookies: AbstractCookieJar, path: str):
    try:
        cookie_jar = unconvert_cookiejar(cookies)
    except Exception as e:
        msg = f'error converting cookie jar: {e}'
        raise CookieStoreError(msg) from e
    try:
        cookie_jar.save(path, ignore_discard=True, ignore_expires=True)
    except Exception as e:
        msg = f'failed to store cookies to "{path}": {e}'
        raise CookieStoreError(msg) from e


def parse_to_timestamp(text: Optional[str]) -> Optional[int]:
    if text is None:
        return None
    try:
        dt = dateutil.parser.parse(str(text))
    except Exception:
        return None
    return int(dt.timestamp())


def unconvert_cookiejar(cookies: AbstractCookieJar) -> cookiejar.MozillaCookieJar:
    cookie_jar = cookiejar.MozillaCookieJar()
    for morsel in cookies:
        domain = morsel.get('domain', '')
        expires = parse_to_timestamp(morsel.get('expires')) or 0
        cookie = http.cookiejar.Cookie(
            version=morsel.get('version') or 0,
            name=morsel.key,
            value=morsel.value,
            port=None,
            port_specified=False,
            domain=domain or '',
            domain_specified=bool(domain),
            domain_initial_dot=domain.startswith('.'),
            path=morsel.get('path', ''),
            path_specified=bool(morsel.get('path')),
            secure=morsel.get('secure') or False,
            expires=expires,
            discard=False,
            comment=morsel.get('comment'),
            comment_url=None,
            rest={},
        )
        cookie_jar.set_cookie(cookie)
    return cookie_jar


def parse_to_date_string(text: Union[int, str, None]) -> Optional[str]:
    if text is None:
        return None
    try:
        dt = dateutil.parser.parse(str(text))
    except Exception:
        return None
    date_string = dt.strftime('%a, %d-%b-%y %H:%M:%S GMT')
    return date_string


def convert_cookiejar(cookie_jar: cookiejar.CookieJar) -> aiohttp.CookieJar:
    """convert cookie jar produced by stdlib to format used by aiohttp"""
    cookies: http.cookies.BaseCookie = http.cookies.BaseCookie()
    for cookie in cookie_jar:
        name = cookie.name
        cookies[name] = cookie.value or ''
        cookies[name]['domain'] = cookie.domain or ''
        cookies[name]['path'] = cookie.path or ''
        cookies[name]['expires'] = parse_to_date_string(cookie.expires) or ''
        cookies[name]['secure'] = cookie.secure or ''
        cookies[name]['version'] = str(cookie.version) if cookie.version else ''
        cookies[name]['comment'] = cookie.comment or ''
    new_jar = aiohttp.CookieJar(quote_cookie=False)
    new_jar.update_cookies(cookies)
    return new_jar


def check_dir(path: Path, create=True) -> bool:
    """check if directory exists and writable, create if asked"""
    if path.is_dir() and os.access(path, mode=os.W_OK):
        return True
    elif create:
        logging.info(f'directory {path} does not exist, creating')
        try:
            path.mkdir(parents=True, exist_ok=True)
            return True
        except OSError as e:
            logging.warning(f'failed to create directory at {path}: {e}')
            return False
    else:
        return False


def parse_timestamp_us(timestamp: Union[str, int, None], ) -> Optional[datetime.datetime]:
    return parse_timestamp(timestamp, 6)


def parse_timestamp_ms(timestamp: Union[str, int, None], ) -> Optional[datetime.datetime]:
    return parse_timestamp(timestamp, 3)


def parse_timestamp(timestamp: Union[str, int, None], fraction: int) -> Optional[datetime.datetime]:
    """parse UNIX timestamp as datetime.datetime"""
    if timestamp is None:
        return None
    try:
        ts = int(timestamp)
        dt = datetime.datetime.fromtimestamp(int(ts / 10 ** fraction), tz=datetime.timezone.utc)
        return dt
    except Exception:
        return None


def show_diff(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> str:
    """pretty-print keys that has different values in dict1 and dict2"""
    keys = {*dict1.keys(), *dict2.keys()}
    diff = []
    for k in keys:
        v1 = str(dict1.get(k, ''))
        repr_v1 = shorten(v1, 60)
        v2 = str(dict2.get(k, ''))
        repr_v2 = shorten(v2, 60)
        if v1 != v2 and json.dumps(v1, sort_keys=True) != json.dumps(v2, sort_keys=True):
            diff.append(f'[{k[:12]:12}]: {repr_v2:60} |->| {repr_v1:60}')
    return '\n'.join(diff)


class timeit(ContextDecorator):
    """measure time call takes, print it in the log"""

    def __init__(self, logger: Optional[logging.Logger] = None):
        self.start: float = 0
        self.end: float = 0
        self.logger = logger

    @property
    def duration(self) -> float:
        return self.end - self.start

    @property
    def timedelta(self) -> datetime.timedelta:
        return datetime.timedelta(seconds=self.duration)

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end = perf_counter()
        if self.logger is not None:
            self.logger.debug(f'took {self.timedelta}')
        return False


class LRUCache:

    def __init__(self, max_size: int = 100):
        if max_size <= 0:
            raise ValueError('Maximum cache size must be a positive integer')
        self._max_size = max_size
        self._data: OrderedDict = OrderedDict()

    def put(self, item: Hashable):
        """Put item in the cache, resize the cache if needed"""
        if not item in self._data:
            self._data[item] = 1
        self._data.move_to_end(item)
        while len(self._data) > self._max_size:
            self._data.popitem(last=False)


def find_matching_field(record: Record, pattern: str, fields: Optional[List[str]] = None) -> Optional[str]:
    name, _ = find_matching_field_name_and_value(record, pattern, fields)
    return name


def find_matching_field_value(record: Record, pattern: str, fields: Optional[List[str]] = None) -> Optional[str]:
    _, value = find_matching_field_name_and_value(record, pattern, fields)
    return value


def find_matching_field_name_and_value(record: Record, pattern: str, fields: Optional[List[str]] = None) -> Tuple[
    Optional[str], Optional[Any]]:
    """
    Return name of the first field of the record that contains pattern,
    return None if nothing found. If fields value specified only check
    fields listed in there.
    """
    for field, value in record:
        if fields is not None and field not in fields:
            continue
        if isinstance(value, Record):
            subrecord_search_result = find_matching_field_name_and_value(value, pattern)
            if subrecord_search_result is not None:
                return subrecord_search_result
        else:
            if str(value).find(pattern) > -1:
                return field, value
    return None, None


def record_has_text(record: Record, text: str) -> bool:
    return find_matching_field(record, text) is not None


def read_file(path: Union[str, Path], encoding=None) -> str:
    """
    Read and return file content in provided encoding

    If decoding file content in provided encoding fails, try again using utf8.
    If it also fails, let the exception propagate. Handling OSError is also
    left to caller.
    """
    with open(path, 'rt', encoding=encoding) as fp:
        try:
            text = fp.read()
            return text
        except UnicodeDecodeError:
            pass
    with open(path, encoding='utf8') as fp:
        text = fp.read()
        return text


def write_file(path: Union[str, Path], content: str, encoding='utf8', backups: int = 0):
    if backups > 0:
        rotate_file(path, depth=backups)
    with open(path, 'wt', encoding=encoding) as fp:
        fp.write(content)


def rotate_file(path: Union[str, Path], depth: int = 10):
    """Move "path" to "path.1", "path.1" to "path.2" and so on down to depth parameter"""
    increment_postfix(path, depth)


def increment_postfix(path: Union[str, Path], maxdepth):
    path = Path(path)
    if not path.exists():
        return
    if re.match(r'\.(\d|[1-9]\d+)$', path.suffix):
        index = int(path.suffix.strip('.'))
        next_path = path.with_suffix(f'.{index + 1}')
    else:
        index = 0
        next_path = path.with_suffix(path.suffix + '.0')
    if index >= maxdepth:
        return
    increment_postfix(next_path, maxdepth)
    logging.getLogger('rotate').info(f'moving {path} to {next_path}')
    path.replace(next_path)


def sha1(text: str) -> str:
    return hashlib.sha1(text.encode()).digest().hex()


def get_cookie_value(jar: Union[CookieJar, AbstractCookieJar], name: str) -> Optional[str]:
    found: List[Union[http.cookiejar.Cookie, http.cookies.Morsel]]
    if isinstance(jar, CookieJar):
        found = [x for x in jar if x.name == name]
    else:
        found = [x for x in jar if x.key == name]
    if not found:
        return None
    return found[0].value


def set_cookie_value(jar: AbstractCookieJar, key: str, value: str, url: str):
    morsel: http.cookies.Morsel = http.cookies.Morsel()
    morsel.set(key, value, value)
    morsel['domain'] = urllib.parse.urlparse(url).netloc
    morsel['path'] = urllib.parse.urlparse(url).path
    jar.update_cookies(morsel)


def find_all(data: JSONType, jsonpath: str, cache={}) -> List[JSONType]:
    if jsonpath not in cache:
        cache[jsonpath] = JSONPath(jsonpath)
    parser = cache[jsonpath]
    return parser.parse(data)


def find_one(data: JSONType, jsonpath: str) -> Optional[JSONType]:
    result = find_all(data, jsonpath)
    return result[0] if result else None


def jwt_decode(token: str) -> dict:
    """Decode JWT token and return payload. Signature is not validated"""
    header, payload, signature = token.split('.')
    payload_json = base64.b64decode(payload.encode('utf-8') + b'====')
    payload_dict = json.loads(payload_json)
    return payload_dict


def utcnow() -> datetime.datetime:
    return datetime.datetime.now(tz=datetime.timezone.utc)


def with_prefix(logger: logging.Logger, prefix: str) -> logging.Logger:
    class Adapter(logging.LoggerAdapter):
        def process(self, msg: Any, kwargs: MutableMapping[str, Any]) -> tuple[Any, MutableMapping[str, Any]]:
            message = f'{prefix} {msg}' if prefix else msg
            return message, kwargs

    return Adapter(logger, extra=dict())  # type: ignore


def is_url(maybe_url: Optional[str]) -> bool:
    if maybe_url is None:
        return False
    try:
        AnyHttpUrl(maybe_url)
        return True
    except ValidationError:
        return False


T = TypeVar('T')


def getitem(container: Dict[str, Any], key: str, expected_type: Type[T]) ->T:
    """
    return container.get(key), raise ValueError if result type doesn't match expected_type
    """
    item = container.get(key)
    if not isinstance(item, expected_type):
        raise ValueError(f'unexpected {key} format: expected {expected_type.__name__}, got {type(item).__name__}')
    return item


class Timezone:
    known: Dict[str, Any] = {}

    @classmethod
    def get_tz(cls, name: Optional[str]) -> Optional[datetime.tzinfo]:
        if name is None:
            return None
        tz = dateutil.tz.gettz(name)
        if tz is None:
            raise ValueError(f'Unknown timezone: {name}')
        cls.known[name] = tz
        return tz

    @classmethod
    def get_name(cls, tz: Optional[datetime.tzinfo]) -> Optional[str]:
        if tz is None:
            return None
        for name, timezone in cls.known.items():
            if tz == timezone:
                return name
        return tz.tzname(datetime.datetime.now())


class DictRootModel(RootModel):
    """Helper class implementing dict methods for dict-based root models"""
    root: dict

    def __getitem__(self, key):
        return self.root[key]

    def __setitem__(self, key, value):
        self.root[key] = value

    def keys(self):
        return self.root.keys()

    def values(self):
        return self.root.values()

    def items(self):
        return self.root.items()


class ListRootModel(RootModel):
    """Helper class implementing indexing and iteration list-based root models"""
    root: list

    def __getitem__(self, index: int):
        return self.root[index]

    def __len__(self) -> int:
        return len(self.root)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.root)

    def __contains__(self, item) -> bool:
        return item in self.root

    def append(self, item):
        self.root.append(item)

    def extend(self, items: Iterable):
        self.root.extend(items)


def strip_text(s: str, text: str) -> str:
    if s.startswith(text):
        return s[len(text):]
    return s


def format_validation_error(e: ValidationError, msg: str) -> str:
    errors = []
    for err in e.errors():
        user_input = str(err['input'])
        user_input = user_input if len(user_input) < 85 else user_input[:50] + ' [...] ' + user_input[-30:]
        location = ': '.join(str(l) for l in err['loc'])
        error_message = strip_text(err['msg'], 'Value error, ')
        error = 'error parsing "{}" in config section {}: {}'
        errors.append(error.format(user_input, location, error_message))
    return '\n    '.join([msg] + errors)



import datetime
import logging
import math
import sqlite3
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple, Union

from pydantic import Field, field_validator, model_validator

from avtdl.core.actors import ActorConfig
from avtdl.core.interfaces import AbstractRecordsStorage, Record, get_record_type
from avtdl.core.utils import check_dir


class BaseRecordDB:
    table_name = 'records'
    table_structure = 'parsed_at datetime, feed_name text, uid text, hashsum text, class_name text, as_json text, PRIMARY KEY(uid, hashsum)'
    row_structure = ':parsed_at, :feed_name, :uid, :hashsum, :class_name, :as_json'
    id_field = 'uid'
    exact_id_field = 'hashsum'
    group_id_field = 'feed_name'
    sorting_field = 'parsed_at'

    def __init__(self, db_path: Union[str, Path], logger: Optional[logging.Logger] = None):
        self.logger = logger or logging.getLogger('RecordDB')
        try:
            if not db_path == ':memory:' and not Path(db_path).exists():
                check_dir(Path(db_path).parent)
            self.db = sqlite3.connect(db_path)
            self.db.row_factory = sqlite3.Row
            self.cursor = self.db.cursor()
            self.cursor.execute('CREATE TABLE IF NOT EXISTS {} ({})'.format(self.table_name, self.table_structure))
            self.db.commit()
        except sqlite3.OperationalError as e:
            self.logger.error(
                f'error opening sqlite database at path "{db_path}", specified in "db_path" config variable: {e}. If file exists make sure it was produced by this application, otherwise check if new file can be created at specified location. Alternatively use special value ":memory:" to use in-memory database instead.')
            raise
        self.logger.debug(f'successfully connected to sqlite database at "{db_path}"')
        self.create_indexes()

    def create_indexes(self):
        queries = [
            f'CREATE INDEX IF NOT EXISTS `index_{self.group_id_field}` ON `{self.table_name}` (`{self.group_id_field}`);',
            f'CREATE INDEX IF NOT EXISTS `index_{self.sorting_field}` ON `{self.table_name}` ( `{self.sorting_field}`)',
            f'CREATE INDEX IF NOT EXISTS `index_{self.id_field}_{self.sorting_field}` ON `{self.table_name}` ( `{self.id_field}`, `{self.sorting_field}` )'
        ]
        for query in queries:
            try:
                self.cursor.execute(query)

            except sqlite3.OperationalError as e:
                self.logger.exception(f'failed to create index: {e}. Raw query: {query}')
        self.db.commit()

    def store(self, rows: Union[Dict[str, Any], List[Dict[str, Any]]], replace: bool = False) -> None:
        on_conflict = 'REPLACE' if replace else 'IGNORE'
        sql = "INSERT OR {} INTO {} VALUES({})".format(on_conflict, self.table_name, self.row_structure)
        if not isinstance(rows, list):
            rows = [rows]
        self.cursor.executemany(sql, rows)
        self.db.commit()

    def fetch_row(self, uid: Any, exact_id: Optional[str] = None) -> Optional[sqlite3.Row]:
        if exact_id is not None:
            sql = f'SELECT * FROM records WHERE {self.id_field}=:uid AND {self.exact_id_field}=:exact_id ORDER BY {self.sorting_field} DESC LIMIT 1'
        else:
            sql = f'SELECT * FROM records WHERE {self.id_field}=:uid ORDER BY {self.sorting_field} DESC LIMIT 1'
        keys = {'uid': uid, 'exact_id': exact_id}
        self.cursor.execute(sql, keys)
        return self.cursor.fetchone()

    def row_exists(self, uid: Any, exact_id: Optional[str] = None) -> bool:
        if exact_id is not None:
            sql = f'SELECT 1 FROM records WHERE {self.id_field}=:uid AND {self.exact_id_field}=:exact_id LIMIT 1'
        else:
            sql = f'SELECT 1 FROM records WHERE {self.id_field}=:uid LIMIT 1'
        keys = {'uid': uid, 'exact_id': exact_id}
        self.cursor.execute(sql, keys)
        return self.cursor.fetchone() is not None

    def get_size(self, group_id: Optional[str] = None) -> int:
        '''return number of records, total or for specified feed, are stored in db'''
        if group_id is None:
            sql = f'SELECT COUNT(1) FROM {self.table_name}'
        else:
            sql = f'SELECT COUNT(1) FROM {self.table_name} WHERE {self.group_id_field}=:group'
        keys = {'group': group_id}
        self.cursor.execute(sql, keys)
        return int(self.cursor.fetchone()[0])

    def get_groups(self) -> List[Tuple[str, int]]:
        sql = f'SELECT {self.group_id_field}, COUNT(1) as count FROM records GROUP BY {self.group_id_field}'
        self.cursor.execute(sql)
        rows = self.cursor.fetchall()
        return [(row[self.group_id_field], int(row['count'])) for row in rows]


    def fetch_offset(self, limit: int, offset: int, group_id: Optional[str] = None, desc: bool = True) -> List[sqlite3.Row]:
        order = 'DESC' if desc else 'ASC'
        if group_id is not None:
            sql = f'SELECT * FROM records WHERE {self.group_id_field}=:group_id ORDER BY {self.sorting_field} {order} LIMIT :limit OFFSET :offset'
        else:
            sql = f'SELECT * FROM records ORDER BY {self.sorting_field} {order} LIMIT :limit OFFSET :offset'
        keys = {'group_id': group_id, 'limit': limit, 'offset': offset}
        self.cursor.execute(sql, keys)
        return self.cursor.fetchall()


def calculate_offset(page: Optional[int], per_page: int, total_rows: int) -> Tuple[int, int]:
    if total_rows == 0 or per_page == 0:
        return 0, 0
    last_page = math.ceil(total_rows / per_page)
    if page is None:
        page = last_page
    if page > last_page:
        page = last_page
    if page < 1:
        page = 1
    page_offset = page * per_page
    offset = max(total_rows - page_offset, 0)
    rows_on_last_page = total_rows % per_page or per_page
    limit = per_page if page != last_page else rows_on_last_page
    return limit, offset


class RecordDB(BaseRecordDB):

    @staticmethod
    def _get_record_id(record: Record, entity_name: str) -> str:
        return '{}:{}'.format(entity_name, record.get_uid())

    def store_records(self, records: Sequence[Record], entity_name: str,
                      replace: bool = False, use_created_as_parsed: bool = False):
        rows = []
        for record in records:
            uid = self._get_record_id(record, entity_name)
            parsed_at = datetime.datetime.now(tz=datetime.timezone.utc)
            if use_created_as_parsed:
                parsed_at = record.created_at
            hashsum = record.hash()
            feed_name = entity_name
            class_name = record.__class__.__name__
            as_json = record.as_json()
            row = {'parsed_at': parsed_at, 'feed_name': feed_name, 'uid': uid, 'hashsum': hashsum,
                   'class_name': class_name, 'as_json': as_json}
            rows.append(row)
        self.store(rows, replace)

    def load_record(self, record: Record, entity_name: str) -> Optional[Record]:
        """load most recently stored version of the record from db"""
        uid = self._get_record_id(record, entity_name)
        stored_record = self.fetch_row(uid)
        if stored_record is None:
            return None
        stored_record_instance = type(record).model_validate_json(stored_record['as_json'])
        return stored_record_instance

    def record_exists(self, record: Record, entity_name: str) -> bool:
        uid = self._get_record_id(record, entity_name)
        return self.row_exists(uid)

    def record_got_updated(self, record: Record, entity_name: str) -> bool:
        """return True when there are different versions of the record in db but not this one"""
        uid = self._get_record_id(record, entity_name)
        return self.row_exists(uid) and not self.row_exists(uid, record.hash())

    def record_has_changed(self, record: Record, entity_name: str, excluded_fields: Set[str]):
        """check if the record differs from most recently stored version, not counting fields listed in excluded_fields"""
        stored_record = self.load_record(record, entity_name)
        if stored_record is None:
            return False
        record_dump = record.model_dump(exclude=excluded_fields)
        stored_record_dump = stored_record.model_dump(exclude=excluded_fields)
        return record_dump != stored_record_dump

    def page_count(self, entity_name: Optional[str], per_page: int) -> int:
        pages = self.get_size(entity_name) / per_page
        pages = math.ceil(pages)
        return pages

    def parse_record(self, row: sqlite3.Row) -> Optional[Record]:
        type_name = row['class_name']
        record_type = get_record_type(type_name)
        if record_type is None:
            self.logger.warning(f'failed to restore record: unsupported record type "{record_type}')
            row_content = {k: row[k] for k in row.keys()}
            self.logger.debug(f'Raw row: {row_content}')
            return None
        record = record_type.model_validate_json(row['as_json'])
        self.set_record_time(record, row['parsed_at'] or None)
        return record

    @staticmethod
    def set_record_time(record: Record, ts: Optional[str]):
        if ts is None:
            return
        try:
            dt = datetime.datetime.fromisoformat(ts)
        except Exception:
            return
        dt = dt.astimezone(tz=datetime.timezone.utc)
        record.created_at = dt

    def load_page(self, entity_name: Optional[str], page: Optional[int], per_page: int, desc: bool = True) -> List[Record]:
        total_rows = self.get_size(entity_name)

        limit, offset = calculate_offset(page, per_page, total_rows)
        rows = self.fetch_offset(limit, offset, entity_name, desc)
        records = []
        for row in rows:
            record = self.parse_record(row)
            if record is not None:
                records.append(record)
        if desc:
            records = records[::-1]
        return records

    def feeds(self) -> List[Tuple[str, int]]:
        return self.get_groups()


class BaseDbConfig(ActorConfig):
    db_path: Union[Path, str] = Field(default='db/', validate_default=True)
    """path to the sqlite database file keeping history of old records.
    Might specify a path to a directory containing the file (with trailing slash)
    or a direct path to the file itself (without a slash). If special value `:memory:` is used,
    database is kept in memory and not stored on disk at all, providing a clean database on every startup"""

    @field_validator('db_path')
    @classmethod
    def str_to_path(cls, path: Union[Path, str]):
        return validate_db_path(path)

    @model_validator(mode='after')
    def handle_db_directory(self):
        if isinstance(self.db_path, Path) and self.db_path.is_dir():
            self.db_path = self.db_path.joinpath(f'{self.name}.sqlite')
        return self


def validate_db_path(path: Union[Path, str]) -> Union[Path, str]:
    if isinstance(path, Path):
        return path
    if path == ':memory:':
        return path
    if path.endswith('/') or path.endswith('\\'):
        ok = check_dir(Path(path), create=True)
        if not ok:
            raise ValueError(f'error accessing path {path}, check if it is a valid path and is writeable')
    return Path(path)


class RecordDbView(AbstractRecordsStorage):

    def __init__(self, db: RecordDB):
        self.db = db

    def feeds(self) -> List[Tuple[str, int]]:
        return self.db.feeds()

    def page_count(self, per_page: int, feed: Optional[str] = None) -> int:
        return self.db.page_count(feed, per_page)

    def load_page(self, page: Optional[int], per_page: int, desc: bool = True, feed: Optional[str] = None) -> List[Record]:
        return self.db.load_page(feed, page, per_page, desc)



import dataclasses
import datetime
from enum import Enum
from typing import Callable, Optional, Sequence, Tuple

from pydantic import Field, PositiveFloat, PositiveInt

from avtdl.core.interfaces import Record
from avtdl.core.monitors import BaseFeedMonitor, BaseFeedMonitorConfig, BaseFeedMonitorEntity
from avtdl.core.plugins import Plugins
from avtdl.core.request import HttpClient
from avtdl.plugins.withny.extractors import WithnyRecord, format_timestamp, parse_live_record, parse_schedule_record, \
    utcnow_with_offset


@Plugins.register('withny', Plugins.kind.ACTOR_CONFIG)
class WithnyMonitorConfig(BaseFeedMonitorConfig):
    pass


class WithnyMonitorMode(str, Enum):
    STREAMS = 'streams'
    SCHEDULES = 'schedules'


@Plugins.register('withny', Plugins.kind.ACTOR_ENTITY)
class WithnyMonitorEntity(BaseFeedMonitorEntity):
    update_interval: PositiveFloat = 300
    """how often the monitored channel will be checked, in seconds"""
    update_ratio: PositiveInt = 12
    """ratio of live to scheduled streams updates"""
    url: str = Field(exclude=True, default='')
    """url is not used, all streams are checked through the same api endpoint"""
    since: Optional[datetime.datetime] = Field(exclude=True, default=None)
    """internal variable, used to store "since" parameter of the last successful update"""
    current_update_ratio: PositiveInt = Field(exclude=True, default=1)
    """internal variable, used to store number of updates of livestreams endpoint since last update of schedules"""


@dataclasses.dataclass
class Context:
    count: int
    page: int
    take: int = 20


@Plugins.register('withny', Plugins.kind.ACTOR)
class WithnyMonitor(BaseFeedMonitor):
    """
    Monitor livestreams on Withny

    Livestreams are checked every update, scheduled streams once every `update_ratio` times.
    Checks all channels at once, so it's recommended to only use a single entity and select
    channels for further processing by using "filter.match" or other filters.
    Does not retrieve playlist_url and therefore does not require authentication.
    """

    async def get_records(self, entity: WithnyMonitorEntity, client: HttpClient) -> Sequence[WithnyRecord]:
        records = await self._get_streams(entity, client)
        if entity.current_update_ratio < entity.update_ratio:
            entity.current_update_ratio += 1
        else:
            entity.current_update_ratio = 1
            upcoming_records, _ = await self._get_schedules(entity, client)
            # deduplicate records, giving the upcoming ones priority
            records = list({record.stream_id: record for record in [*records, *upcoming_records]}.values())
        return records

    async def _parse_data(self, data, parser: Callable) -> Sequence[WithnyRecord]:
        records = []
        for item in data:
            try:
                record = parser(item)
                records.append(record)
            except Exception as e:
                self.logger.exception(f'failed to parse stream record: {e}')
                self.logger.debug(f'raw record data: {item}')
        return records

    async def _get_streams(self, entity: WithnyMonitorEntity, client: HttpClient) -> Sequence[WithnyRecord]:
        url = 'https://www.withny.fun/api/streams/with-rooms'
        data = await self.request_json(url, entity, client)
        if data is None:
            return []
        if not isinstance(data, list):
            self.logger.warning(f'unexpected data from /streams/ api: stream list is not a list')
            self.logger.debug(f'raw /streams/ api response: {data}')
            return []
        records = await self._parse_data(data, parse_live_record)
        return records

    async def _get_schedules(self,
                             entity: WithnyMonitorEntity,
                             client: HttpClient,
                             context: Optional[Context] = None) -> Tuple[Sequence[WithnyRecord], Context]:
        url = 'https://www.withny.fun/api/schedules'
        context = context or Context(count=0, page=1)

        if entity.since is None:
            entity.since = utcnow_with_offset(hours=-3)
        params = {
            'page': context.page,
            'take': context.take,
            'isFavorite': 'false',
            'excludeClosedStream': 'false',
            'later': format_timestamp(entity.since)
        }
        data = await self.request_json(url, entity, client, params=params)
        if data is None or not isinstance(data, dict):
            return [], context

        for name, _type in [('schedules', list), ('count', int)]:
            if name not in data or not isinstance(data[name], _type):
                self.logger.warning(f'unexpected data from /schedules/ api')
                self.logger.debug(f'raw /schedules/ api response: {data}')
                return [], context

        schedules_data = data['schedules']
        records = await self._parse_data(schedules_data, parse_schedule_record)

        context.count = data['count']
        self.logger.debug(f'[{entity.name}] schedules: {len(records)}/{context.count} records on page {context.page}')
        if context.page * context.take > context.count:  # this page is the last one
            return records, context
        if context.page < 2 and not all([self.record_is_new(record, entity) for record in records]):
            self.logger.debug(f'found already seen records on page {context.page}, aborting early')
            return records, context

        context.page += 1
        next_page_records, _ = await self._get_schedules(entity, client, context)
        records = [*records, *next_page_records]

        return records, context

    async def get_new_records(self, entity: WithnyMonitorEntity, client: HttpClient) -> Sequence[Record]:
        new_records = await super().get_new_records(entity, client)
        if new_records:
            # updating "since" invalidates caching headers, so it's only done when response content has changed anyway
            entity.since = utcnow_with_offset(hours=-3)
            self.logger.debug(f'[{entity.name}] got {len(new_records)} new records, "since" set to {entity.since}')
        return new_records



import asyncio
import logging
import signal
from collections import defaultdict, deque
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Coroutine, Dict, List, Literal, Optional, Tuple

from pydantic import Field

from avtdl.core.interfaces import Record
from avtdl.core.state import StateSerializer
from avtdl.core.utils import DictRootModel

Subscription = Callable[[str, Record], None]
SubscriptionsMapping = Dict[str, List[Subscription]]

HISTORY_SIZE = 20


def deque_factory() -> deque:
    return deque(maxlen=HISTORY_SIZE)


class MessageHistory(DictRootModel):
    root: Dict[str, deque[Record]] = Field(default_factory=lambda: defaultdict(deque_factory))


class MessageBus:
    PREFIX_IN = 'inputs'
    PREFIX_OUT = 'output'
    SEPARATOR = '/'

    PERSISTENCE_FILE = Path('bus/history.dat')

    def __init__(self) -> None:
        self.subscriptions: SubscriptionsMapping = defaultdict(list)
        self.logger = logging.getLogger('bus')
        self.history: MessageHistory = MessageHistory()

    def sub(self, topic: str, callback: Subscription):
        self.logger.debug(f'subscription on topic {topic} by {callback!r}')
        self.subscriptions[topic].append(callback)

    def _generic_topic(self, specific_topic: str) -> str:
        direction, actor, entity, chain = self.split_subscription_topic(specific_topic)
        generic_topic = self.make_topic(direction, actor, entity, '')
        return generic_topic

    def pub(self, topic: str, message: Record):
        self.logger.debug(f'on topic {topic} message "{message!r}"')
        matching_callbacks = self.get_matching_callbacks(topic)
        for specific_topic, callbacks in matching_callbacks.items():
            if message.chain:
                targeted_message = message
            else:
                _, _, chain = self.split_message_topic(specific_topic)
                targeted_message = message.model_copy(deep=True)
                targeted_message.chain = chain
            for callback in callbacks:
                callback(specific_topic, targeted_message)

            generic_topic = self._generic_topic(specific_topic)
            self.add_to_history(generic_topic, targeted_message)
        if not matching_callbacks:
            # topic has no subscribers, meaning entity is not referenced in chains
            self.add_to_history(self._generic_topic(topic), message)

    def add_to_history(self, topic: str, message: Record):
        self.history[topic].append(message)

    def get_history(self, actor: str, entity: str, chain: str = '', direction: Literal['in', 'out'] = 'in') -> List[Record]:
        if direction == 'in':
            topic = self.incoming_topic_for(actor, entity, '')
        elif direction == 'out':
            topic = self.outgoing_topic_for(actor, entity, '')
        else:
            assert False, f'unexpected direction "{direction}"'
        if chain:
            records = [record for record in self.history[topic] if record.chain == chain]
        else:
            records = list(self.history[topic])
        records.sort(key = lambda record: record.created_at)
        return records

    def get_matching_callbacks(self, topic_pattern: str) -> SubscriptionsMapping:
        pattern_direction, pattern_actor, pattern_entity, pattern_chain = self.split_subscription_topic(topic_pattern)
        callbacks = defaultdict(list)
        for topic, callback in self.subscriptions.items():
            direction, actor, entity, chain = self.split_subscription_topic(topic)
            if pattern_direction != direction:
                continue
            if actor != pattern_actor:
                continue
            if entity != pattern_entity:
                continue
            if chain == '' or pattern_chain == '' or chain == pattern_chain:
                callbacks[topic].extend(callback)
        return callbacks

    def make_topic(self, *args: str):
        return self.SEPARATOR.join(args)

    def split_topic(self, topic: str):
        return topic.split(self.SEPARATOR)

    def incoming_topic_for(self, actor: str, entity: str, chain: str = '') -> str:
        return self.make_topic(self.PREFIX_IN, actor, entity, chain)

    def outgoing_topic_for(self, actor: str, entity: str, chain: str = '') -> str:
        return self.make_topic(self.PREFIX_OUT, actor, entity, chain)

    def split_message_topic(self, topic) -> Tuple[str, str, str]:
        _, actor, entity, chain = self.split_subscription_topic(topic)
        return actor, entity, chain

    def split_subscription_topic(self, topic) -> Tuple[str, str, str, str]:
        try:
            direction, actor, entity, chain = self.split_topic(topic)
        except ValueError:
            self.logger.error(f'failed to split message topic "{topic}"')
            raise
        return direction, actor, entity, chain

    def clear_subscriptions(self):
        self.subscriptions.clear()

    def dump_state(self, directory: Path):
        StateSerializer.dump(self.history, directory / self.PERSISTENCE_FILE)

    def apply_state(self, directory: Path):
        stored_history = StateSerializer.restore(MessageHistory, directory / self.PERSISTENCE_FILE)
        if stored_history is None:
            return
        for topic, stored_topic_history in stored_history.items():
            active_topic_history = self.history[topic]
            for record in reversed(stored_topic_history):
                if len(active_topic_history) >= (active_topic_history.maxlen or 100500):
                    break
                active_topic_history.appendleft(record)


class TerminatedAction(int, Enum):
    EXIT = 0
    RESTART = 2


@dataclass
class TaskStatus:
    actor: Optional[str]
    entity: Optional[str]
    status: str = ''
    record: Optional[Record] = None

    def set_status(self, status: str, record: Optional[Record] = ...):  # type: ignore
        self.status = status
        if record is not ...:
            self.record = record

    def clear(self):
        self.status = ''
        self.record = None

    def is_empty(self) -> bool:
        return not self.status and not self.record


class TasksController:
    class TerminatedError(KeyboardInterrupt):
        """Raised when application restart is requested"""

    def __init__(self, poll_interval: float = 5, logger: Optional[logging.Logger] = None) -> None:
        self.logger = logger or logging.getLogger('task_controller')
        self.poll_interval = poll_interval
        self.termination_pending = False
        self.termination_required = False
        self.terminated_action = TerminatedAction.EXIT
        self.tasks: set[asyncio.Task] = set()
        self._info: Dict[asyncio.Task, Optional[TaskStatus]] = {}

    def create_task(self, coro: Coroutine, *, name: Optional[str] = None,
                    _info: Optional[TaskStatus] = None) -> asyncio.Task:
        task = asyncio.create_task(coro, name=name)
        if task in self.tasks:
            raise RuntimeError(f'newly created task {task} is already monitored')
        self.tasks.add(task)
        self._info[task] = _info
        return task

    async def check_done_tasks(self, done: set[asyncio.Task]) -> None:
        for task in done:
            if not task.done():
                continue
            try:
                task_exception = task.exception()
                if task_exception is not None:
                    self.logger.error(f'task "{task.get_name()}" has terminated with exception', exc_info=task_exception)
            except asyncio.CancelledError:
                self.logger.debug(f'task "{task.get_name()}" cancelled')
            self.tasks.discard(task)
            self._info.pop(task)

    async def monitor_tasks(self) -> None:
        while not self.termination_required:
            if not self.tasks:
                await asyncio.sleep(self.poll_interval)
                continue
            done, pending = await asyncio.wait(self.tasks, return_when=asyncio.FIRST_EXCEPTION, timeout=self.poll_interval)
            await self.check_done_tasks(done)
        self.termination_required = False
        raise self.TerminatedError()

    async def cancel_all_tasks(self) -> None:
        self.logger.debug(f'terminating {len(self.tasks)} tasks')
        while True:
            for task in self.tasks:
                task.cancel('terminating')
            done, pending = await asyncio.wait(self.tasks, return_when=asyncio.ALL_COMPLETED)
            await self.check_done_tasks(done)
            if not pending:
                break
            self.logger.debug(f'{len(pending)} more tasks left to terminate')
        self.logger.debug('all tasks terminated')

    async def terminate(self, delay: float, action: TerminatedAction):
        if self.termination_pending:
            self.logger.warning(f'active termination request is already pending')
            return
        self.termination_pending = True
        if delay > 0:
            self.logger.debug(f'terminating after {delay:.02f}')
            await asyncio.sleep(delay)
        self.logger.debug(f'terminating now')
        self.termination_pending = False
        self.terminated_action = action
        self.termination_required = True

    def terminate_after(self, delay: float, action: TerminatedAction):
        self.create_task(self.terminate(delay, action), name=f'terminate after {delay}')

    async def run_until_termination(self) -> TerminatedAction:
        try:
            await self.monitor_tasks()
        except self.TerminatedError:
            await self.cancel_all_tasks()
        except (asyncio.CancelledError, KeyboardInterrupt):
            await self.cancel_all_tasks()
            raise
        return self.terminated_action

    def get_status(self) -> List[TaskStatus]:
        return [status for status in self._info.values() if status is not None]

    def get_task_status(self, task_name: str) -> Optional[TaskStatus]:
        for task, task_info in self._info.items():
            if task.get_name() == task_name:
                return task_info
        return None

class RuntimeContext:
    def __init__(self, bus: MessageBus, controller: TasksController):
        self.bus: MessageBus = bus
        self.controller: TasksController = controller
        self.extra: Dict[str, Any] = {}
        self._sigint_handler = signal.getsignal(signal.SIGINT)
        self._sigterm_handler = signal.getsignal(signal.SIGINT)

    def set_extra(self, name: str, value: Any):
        """
        Store extra object in RuntimeContext

        Used to add objects that doesn't exist when RuntimeContext is initialized
        (namely Settings instance)
        """
        self.extra[name] = value

    def get_extra(self, name: str) -> Optional[Any]:
        """Retrieve object stored by `set_extra`"""
        return self.extra.get(name)

    def _get_handler(self) -> Callable:
        controller = self.controller
        def handler(sig, frame):
            controller.terminate_after(0, TerminatedAction.EXIT)
        return handler

    def __enter__(self):
        signal.signal(signal.SIGINT, self._get_handler())
        signal.signal(signal.SIGTERM, self._get_handler())

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.signal(signal.SIGINT, self._sigint_handler)
        signal.signal(signal.SIGTERM, self._sigterm_handler)


    @classmethod
    def create(cls) -> 'RuntimeContext':
        bus = MessageBus()
        controller = TasksController()
        return cls(bus=bus, controller=controller)



from pathlib import Path
from typing import Optional, Sequence, Union

from pydantic import Field, field_validator, model_validator

from avtdl.core.actors import Action, ActionEntity, ActorConfig
from avtdl.core.db import RecordDB, RecordDbView, validate_db_path
from avtdl.core.interfaces import AbstractRecordsStorage, Record
from avtdl.core.plugins import Plugins
from avtdl.core.runtime import RuntimeContext

Plugins.register('view', Plugins.kind.ASSOCIATED_RECORD)(Record)


@Plugins.register('view', Plugins.kind.ACTOR_CONFIG)
class ViewConfig(ActorConfig):
    pass


@Plugins.register('view', Plugins.kind.ACTOR_ENTITY)
class ViewEntity(ActionEntity):
    db_path: Union[Path, str] = Field(default='db/', validate_default=True)
    """path to the sqlite database file storing records.
    Might specify a path to a directory containing the file (with trailing slash)
    or a direct path to the file itself (without a slash). If special value `:memory:` is used,
    database is kept in memory and not stored on disk at all, providing a clean database on every startup"""
    readonly: bool = False
    """when enabled, prevents any writes to the database"""
    replace: bool = True
    """when exactly the same record is received more than once, the latest copy will overwrite already stored one,
    updating the time the record was stored"""
    use_creation_timestamp: bool = False
    """use time record was initially parsed at instead of the moment it was stored in the database for chronological sorting"""

    @field_validator('db_path')
    @classmethod
    def str_to_path(cls, path: Union[Path, str]):
        return validate_db_path(path)

    @model_validator(mode='after')
    def handle_db_directory(self):
        if isinstance(self.db_path, Path) and self.db_path.is_dir():
            self.db_path = self.db_path.joinpath(f'view/{self.name}.sqlite')
        return self


@Plugins.register('view', Plugins.kind.ACTOR)
class View(Action):
    """
    View records in the web interface

    Incoming records are stored in sqlite database at db_path (unless readonly parameter enabled).
    The stored records are available on the "View" tab of the web interface.

    This plugin is under active development. The way it works might change, options might be
    renamed or removed.
    """

    def __init__(self, conf: ViewConfig, entities: Sequence[ViewEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.databases = {}
        for entity in entities:
            db = RecordDB(entity.db_path, logger=self.logger.getChild('name'))
            self.databases[entity.name] = db

    def handle(self, entity: ViewEntity, record: Record):
        if entity.readonly:
            self.logger.debug(f'[{entity.name}] readonly mode, the following record will not be stored: {record!r}')
            return
        db = self.databases.get(entity.name)
        if db is None:
            self.logger.exception(
                f'no database is opened for entity {entity.name}, the following record will not be stored: {record!r}')
            return
        if record.origin:
            origin_items = record.origin.split(':', maxsplit=1)
            entity_name = origin_items[-1]  # take entire origin value if no separator present
        else:
            entity_name = entity.name
        db.store_records([record], entity_name, entity.replace, use_created_as_parsed=entity.use_creation_timestamp)

    def get_records_storage(self, entity_name: Optional[str] = None) -> Optional[AbstractRecordsStorage]:
        if entity_name not in self.databases:
            return None
        return RecordDbView(self.databases[entity_name])



from pathlib import Path
from typing import List, Optional

import pytest

from avtdl.core.cache import FileCache, find_file, find_free_suffix, find_with_suffix, strip_rename_suffix

SUFFIX_TEMPLATE = FileCache.RENAME_SUFFIX


class TestStripRenameSuffix:
    testcases = {
        'no suffix': ('path/to/filename', 'path/to/filename'),
        'one suffix': ('path/to/filename', 'path/to/filename [2]'),
        'suffix is zero': ('path/to/filename', 'path/to/filename [0]'),
        'multiple suffixes': ('path/to [1]/longer [2] filename [3]', 'path/to [1]/longer [2] filename [3] [4]')
    }

    @pytest.mark.parametrize('path, path_with_suffix', testcases.values(), ids=testcases.keys())
    def test_strip_suffix(self, path, path_with_suffix):
        path = Path(path)
        path_with_suffix = Path(path_with_suffix)
        new_name = strip_rename_suffix(path_with_suffix.stem, SUFFIX_TEMPLATE)
        result = path_with_suffix.with_stem(new_name)
        assert result == path


def touch(path: Path, name: str):
    full_name = path / name
    with full_name.open('a'):
        pass
    return full_name


def prepare_files(workdir: Path, files: List[str]) -> List[Path]:
    """create empty files with given names in workdir"""
    return [touch(workdir, file) for file in files]


class TestFindFile:
    testcases = {
        # query, expected results, other files
        'no files': ('file1', [], []),
        'no matching files': ('file1', [], ['file2.jpg']),
        'one matching file': ('file1', ['file1.jpg'], ['file2.jpg', 'file3.jpg']),
        'two matching files': ('file1', ['file1.jpg', 'file1.exe'], ['file2.jpg', 'file3.jpg']),
    }

    @pytest.mark.parametrize('query, expected, unexpected', testcases.values(), ids=testcases.keys())
    def test_find_file(self, tmp_path, query: str, expected: List[str], unexpected: List[str]):
        _ = prepare_files(tmp_path, unexpected)
        expected_paths = prepare_files(tmp_path, expected)

        result = find_file(tmp_path / query)
        assert sorted(result) == sorted(expected_paths)


class TestFindFreeSuffix:
    testcases = {
        # query, expected filename, existing files
        'no files': ('file1', 'file1', []),
        'no matching files': ('file1', 'file1', ['file2.jpg']),
        'file exists': ('file1', 'file1 [1]', ['file1.jpg', 'file2.jpg']),
        'file with index exists': ('file1', 'file1 [2]', ['file1.jpg', 'file1 [1].jpg']),
        'gap in index': ('file1', 'file1 [2]', ['file1.jpg', 'file1 [1].jpg', 'file1 [3].jpg']),
        'no file, but file with index exists': ('file1', 'file1', ['file1 [0].jpg', 'file1 [1].jpg']),
    }

    @pytest.mark.parametrize('query, expected, files', testcases.values(), ids=testcases.keys())
    def test_find_free_suffix(self, tmp_path, query: str, expected: str, files: List[str]):
        prepare_files(tmp_path, files)

        result = find_free_suffix(tmp_path / query, SUFFIX_TEMPLATE)
        assert result.name == expected


class TestFindWithSuffix:
    testcases = {
        # query, expected filename, existing files
        'no files': ('file1', [], []),
        'no matching files': ('file1', [], ['file2.jpg']),
        'file exists': ('file1', ['file1.jpg'], ['file1.jpg', 'file2.jpg']),
        'two files exists': ('file1', ['file1.jpg', 'file1.exe'], ['file1.jpg', 'file1.exe', 'file2.png']),
        'file with index exists': ('file1', ['file1.jpg', 'file1 [1].jpg'], ['file1.jpg', 'file1 [1].jpg']),
        'no file, but file with index exists': ('file1',
                                                ['file1 [0].jpg', 'file1 [1].jpg'],
                                                ['file1 [0].jpg', 'file1 [1].jpg', 'file2.jpg', 'file2 [1].jpg']),
        'gap in index': ('file1',
                         ['file1.jpg', 'file1 [1].jpg', 'file1 [3].jpg'],
                         ['file1.jpg', 'file1 [1].jpg', 'file1 [3].jpg']),
    }

    @pytest.mark.parametrize('query, expected, files', testcases.values(), ids=testcases.keys())
    def test_find_free_suffix(self, tmp_path, query: str, expected: List[str], files: List[str]):
        prepare_files(tmp_path, files)

        results = find_with_suffix(tmp_path / query, SUFFIX_TEMPLATE)
        results_names = [result.name for result in results]
        assert sorted(results_names) == sorted(expected)


def file_cache(tmp_path):
    return FileCache(tmp_path, '.part')


class TestFileCacheFindFile:
    QUERY = 'file1'
    URL = 'http://example.com/file1.jpg'

    testcases = {
        # expected result filename, existing files
        'no files': (None, []),
        'no matching files': (None, ['file2.jpg']),
        'one matching file': ('file1.jpg', ['file1.jpg', 'file2.jpg', 'file3.jpg']),
        'same extension preferred for two matching files': ('file1.jpg',
                                                            ['file1.exe', 'file2.jpg', 'file3.jpg']),
        'higher index preferred for two matching files': ('file1 [3].jpg',
                                                          ['file1.jpg', 'file1 [1].jpg', 'file1 [3].jpg']),
    }

    @pytest.mark.parametrize('expected, files', testcases.values(), ids=testcases.keys())
    def test_find_file(self, tmp_path, expected: Optional[str], files: List[str]):
        prepare_files(tmp_path, files)
        expected = touch(tmp_path, expected) if expected else None

        result = file_cache(tmp_path)._find_file(tmp_path / self.QUERY, self.URL)
        assert result == expected



import datetime
import logging
from abc import ABC, abstractmethod
from textwrap import shorten
from typing import Any, Dict, Optional, Sequence

from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator

from avtdl.core.interfaces import AbstractRecordsStorage, Event, MAX_REPR_LEN, Record
from avtdl.core.runtime import RuntimeContext
from avtdl.core.utils import Timezone


class ActorConfig(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)

    name: str
    defaults: Dict[str, Any] = Field(default={}, exclude=True)


class ActorEntity(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)

    name: str
    """name of a specific entity. Used to reference it in `chains` section. Must be unique within a plugin"""
    reset_origin: bool = False
    """treat throughput records as if they have originated from this entity, 
    emitting them in every Chain this entity is used"""


class Actor(ABC):
    model_config = ConfigDict(use_attribute_docstrings=True)

    def __init__(self, conf: ActorConfig, entities: Sequence[ActorEntity], ctx: RuntimeContext):
        self.conf = conf
        self.logger = logging.getLogger(f'actor').getChild(conf.name)
        self.ctx = ctx
        self.bus = ctx.bus
        self.controller = ctx.controller
        self.entities = {entity.name: entity for entity in entities}

        for entity_name in self.entities:
            topic = self.bus.incoming_topic_for(self.conf.name, entity_name)
            self.bus.sub(topic, self._handle)

    def __repr__(self):
        text = f'{self.__class__.__name__}({list(self.entities)})'
        return shorten(text, MAX_REPR_LEN)

    def _handle(self, topic: str, record: Record) -> None:
        _, entity_name, _ = self.bus.split_message_topic(topic)
        if entity_name not in self.entities:
            logging.warning(f'received record on topic {topic}, but have no entity with name {entity_name} configured, dropping record {record!r}')
            return
        entity = self.entities[entity_name]
        try:
            self.handle_record(entity, record)
        except Exception:
            self.logger.exception(f'{self.conf.name}.{entity_name}: error while processing record "{record!r}"')

    def get_records_storage(self, entity_name: Optional[str] = None) -> Optional[AbstractRecordsStorage]:
        '''
        Implementations might overwrite this method to give web interface access to persistent records storage
        If implementation uses a single storage for all entities, it must return it with entity_name=None,
        and return None when entity_name provided. If each entity uses individual storage, it must be returned
        when entity_name is specified, and for entity_name=None the return value should be None
        '''
        return None

    @abstractmethod
    def handle_record(self, entity: ActorEntity, record: Record) -> None:
        '''Perform an action on record if entity in self.entities'''

    def on_record(self, entity: ActorEntity, record: Record):
        '''Implementation should call it for every new Record it produces'''
        origin = f'{self.conf.name}:{entity.name}'
        if record.origin == origin:
            self.logger.warning(f'[{entity.name}] received incoming record produced by self, which indicates loop in a chain, dropping: "{record!r}".\nCheck config for chains, passing records to each other in a loop. Record has chain set to "{record.chain}"')
            return
        if record.origin is None:
            record.origin = origin
        if entity.reset_origin:
            record = record.model_copy(deep=True)
            record.chain = ''
        topic = self.bus.outgoing_topic_for(self.conf.name, entity.name, record.chain)
        self.bus.pub(topic, record)

    async def run(self):
        '''Will be run as asyncio task once everything is set up'''
        return


class MonitorEntity(ActorEntity):
    reset_origin: bool = Field(default=False, exclude=True)
    """excluded for Monitors because no good usecases"""


class Monitor(Actor, ABC):

    def __init__(self, conf: ActorConfig, entities: Sequence[MonitorEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

    def handle_record(self, entity: MonitorEntity, record: Record) -> None:
        # For convenience’s sake monitors pass incoming records down the chain.
        # It allows using multiple monitors in a single chain, one after another
        self.on_record(entity, record)

    def on_record(self, entity: ActorEntity, record: Record):
        '''Implementation should call it for every new Record it produces'''
        super().on_record(entity, record)


class FilterEntity(ActorEntity):
    pass


class Filter(Actor):

    def __init__(self, conf: ActorConfig, entities: Sequence[FilterEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.logger = logging.getLogger(f'filters.{self.conf.name}')

    def handle_record(self, entity: FilterEntity, record: Record):
        filtered = self.match(entity, record)
        if filtered is not None:
            filtered.origin = filtered.origin or record.origin
            if filtered.chain is None:
                filtered.chain = ''
            elif not filtered.chain:
                filtered.chain = record.chain
            self.on_record(entity, filtered)
        else:
            self.logger.debug(f'[{entity.name}] record dropped: "{record!r}"')

    @abstractmethod
    def match(self, entity: FilterEntity, record: Record) -> Optional[Record]:
        '''
        Take a record and return it or a new/updated record
        if it matches some condition, otherwise return None.

        If returned record does not have "origin" field, it
        is copied from the original one. "chain" field is
        also copied if not set unless it is set to None, in
        which case it is set to empty string, making record
        propagate in all chains the filter is registered in.
        '''


class ActionEntity(ActorEntity):
    consume_record: bool = True
    """whether record should be consumed or passed down the chain after processing. Disabling it allows chaining multiple Actions"""
    event_passthrough: bool = False
    """whether events should be treated as regular records. When enabled, events are passed down the chain without processing, unless `consume_record` is also enabled"""
    timezone: Optional[datetime.tzinfo] = None
    """takes timezone name from <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones> (or local time if omitted), converts record fields containing date and time to this timezone"""

    @field_validator('timezone', mode='plain')
    @classmethod
    def check_timezone(cls, timezone: Optional[str]) -> Optional[datetime.tzinfo]:
        if timezone is None:
            return None
        if not isinstance(timezone, str):
            raise ValueError('Input should be a valid string')
        return Timezone.get_tz(timezone)

    @field_serializer('timezone')
    @classmethod
    def serialize_timezone(cls, timezone: Optional[datetime.tzinfo]) -> Optional[str]:
        return Timezone.get_name(timezone)


class Action(Actor, ABC):

    def __init__(self, conf: ActorConfig, entities: Sequence[ActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

    def handle_record(self, entity: ActionEntity, record: Record) -> None:
        record = record.as_timezone(entity.timezone)
        if not entity.event_passthrough or not isinstance(record, Event):
            self.handle(entity, record)
        if not entity.consume_record:
            self.on_record(entity, record)

    @abstractmethod
    def handle(self, entity: ActionEntity, record: Record):
        '''Method for implementation to process incoming record'''



import asyncio
import datetime
from typing import Sequence, Union

from pydantic import field_serializer, field_validator

from avtdl.core.actors import Filter, FilterEntity
from avtdl.core.interfaces import Record
from avtdl.core.plugins import Plugins
from avtdl.core.runtime import RuntimeContext
from avtdl.plugins.filters.filters import EmptyFilterConfig
from avtdl.plugins.youtube.youtube_feed import YoutubeVideoRecord
from avtdl.plugins.youtube.youtube_rss import YoutubeFeedRecord

Plugins.register('filter.channel.notify', Plugins.kind.ASSOCIATED_RECORD)(YoutubeVideoRecord)
Plugins.register('filter.channel.notify', Plugins.kind.ASSOCIATED_RECORD)(YoutubeFeedRecord)


@Plugins.register('filter.channel.notify', Plugins.kind.ACTOR_CONFIG)
class ChannelNotifyFilterConfig(EmptyFilterConfig):
    pass


@Plugins.register('filter.channel.notify', Plugins.kind.ACTOR_ENTITY)
class ChannelNotifyFilterEntity(FilterEntity):
    prior: Union[int, datetime.timedelta] = 10
    """output a record this many minutes before the scheduled start of a live broadcast"""
    include_ongoing: bool = False
    """whether currently live streams should be included"""

    @field_validator('prior')
    @classmethod
    def to_timedelta(cls, minutes: Union[int, datetime.timedelta]) -> datetime.timedelta:
        if isinstance(minutes, datetime.timedelta):
            return minutes
        return datetime.timedelta(minutes=minutes)

    @field_serializer('prior')
    def from_timedelta(self, minutes: datetime.timedelta, _info) -> int:
        return int(minutes.total_seconds() / 60)


@Plugins.register('filter.channel.notify', Plugins.kind.ACTOR)
class ChannelNotifyFilter(Filter):
    """
    Hold upcoming stream's records until the start time

    Determines whether a record represents a Youtube livestream/Premiere
    with a scheduled time, and holds it waiting until the time comes
    instead of passing down the chain immediately if needed.

    If the record is not an upcoming Youtube livestream, it gets silently dropped.
    """

    def __init__(self, config: ChannelNotifyFilterConfig, entities: Sequence[ChannelNotifyFilterEntity], ctx: RuntimeContext):
        super().__init__(config, entities, ctx)

    def match(self, entity: ChannelNotifyFilterEntity, record: Record) -> None:
        if isinstance(record, YoutubeVideoRecord):
            scheduled = record.scheduled
            if entity.include_ongoing and record.is_live:
                scheduled = datetime.datetime.now(datetime.timezone.utc)
        else:
            scheduled = getattr(record, 'scheduled', ...)  # type: ignore
        if scheduled is ...:
            self.logger.debug(f'[{entity.name}] record has no "scheduled" field, dropping: {record!r}')
            return
        if scheduled is None:
            self.logger.debug(f'[{entity.name}] "scheduled" field of the record is empty, dropping: {record!r}')
            return
        if not isinstance(scheduled, datetime.datetime):
            self.logger.debug(f'[{entity.name}] "scheduled" field of the record has unexpected value, dropping: {record!r}')
            return

        prior = datetime.timedelta(seconds=entity.prior) if isinstance(entity.prior, int) else entity.prior
        at = scheduled - prior
        self.controller.create_task(self.notify(at, entity, record))

    async def notify(self, at: datetime.datetime, entity: ChannelNotifyFilterEntity, record: Record):
        now = datetime.datetime.now(datetime.timezone.utc)
        delay = (at - now).total_seconds()
        if delay > 0:
            self.logger.debug(f'[{entity.name}] waiting {at - now} before emitting {record!r}')
            await asyncio.sleep(delay)
        else:
            self.logger.debug(f'[{entity.name}] deadline was {now - at} ago, emitting record immediately: {record!r}')
        self.on_record(entity, record)



from pathlib import Path
from typing import Any, Dict, List, Optional

import pytest

from avtdl.core.actors import ActorConfig
from avtdl.core.interfaces import TextRecord
from avtdl.core.runtime import RuntimeContext
from avtdl.plugins.file.text_file import FileAction, FileActionEntity


def read_file(path: Path, encoding: Optional[str] = None) -> str:
    with open(path, 'r+t', encoding=encoding) as fp:
        return fp.read()


@pytest.fixture()
def params():
    """provides default value of no additional parameters, replaced in parametrized classes"""
    return {}


@pytest.fixture()
def entity(tmp_path, params) -> FileActionEntity:
    entity_params = {'name': 'test_entity', 'path': tmp_path,  'filename': 'data.txt'}
    entity_params.update(params)
    entity = FileActionEntity.model_validate(entity_params)
    return entity


@pytest.fixture()
def actor(entity) -> FileAction:
    ctx = RuntimeContext.create()
    actor = FileAction(ActorConfig(name='to_file'), [entity], ctx)
    return actor


class TestFirstUpdate:
    testcases = {
        'default entity settings':
            (
                {},
                ['Line 1\nLine 2'],
                'Line 1\nLine 2\n'
            ),
        'output format set to text':
            (
                {'output_format': 'text'},
                ['Line 1', 'Line 2'],
                'Line 1\nLine 2\n'
            ),
        'overwrite with append: both records are written':
            (
                {'overwrite': True, 'append': True},
                ['Line 1', 'Line 2'],
                'Line 1\nLine 2\n'
            ),
        'overwrite with no append: entire file is overwritten':
            (
                {'overwrite': True, 'append': False},
                ['Line 1', 'Line 2'],
                'Line 2\n'
            ),
        'no overwrite with append: only first record written':
            (
                {'overwrite': False, 'append': True},
                ['Line 1', 'Line 2'],
                'Line 1\n'
            ),
        'no overwrite with no append: only first record written':
            (
                {'overwrite': False, 'append': False},
                ['Line 1', 'Line 2'],
                'Line 1\n'
            ),
        'format with missing placeholder':
            (
                {'output_template': '{text} : {image}', 'missing': 'N/A'},
                ['Line 1'],
                'Line 1 : N/A\n'
            ),
        'format without missing placeholder':
            (
                {'output_template': '{text} : {image}'},
                ['Line 1'],
                'Line 1 : {image}\n'
            ),
        'prefix and postfix':
            (
                {'prefix': '---', 'postfix': '|||'},
                ['Line 1', 'Line 2'],
                '---Line 1|||---Line 2|||'
            ),
    }

    @staticmethod
    @pytest.mark.asyncio
    @pytest.mark.parametrize('params, data, expected', testcases.values(), ids=testcases.keys())
    async def test_single_update(entity: FileActionEntity, actor: FileAction, params: Dict[str, Any], data: List[str], expected: str):
        records = [TextRecord(text=text) for text in data]
        for record in records:
            actor.handle(entity, record)
        path = (entity.path or Path()) / entity.filename
        output = read_file(path, encoding=entity.encoding)
        assert output == expected



import dataclasses
import datetime
from typing import Any, Dict, List, Optional

import dateutil.parser

from avtdl.core.interfaces import Record
from avtdl.core.plugins import Plugins
from avtdl.core.utils import utcnow


@Plugins.register('withny', Plugins.kind.ASSOCIATED_RECORD)
class WithnyRecord(Record):
    """
    Ongoing or upcoming livestream on Withny
    """
    url: str
    """channel url"""
    stream_id: str
    """unique id of the stream"""
    title: str
    """stream title"""
    description: str
    """stream description"""
    thumbnail_url: Optional[str]
    """url of the stream thumbnail"""
    start: Optional[datetime.datetime]
    """time of the stream start, if started"""
    end: Optional[datetime.datetime] = None
    """time of the stream end, if ended"""
    scheduled: Optional[datetime.datetime] = None
    """scheduled date for the upcoming stream"""
    schedule_id: Optional[str] = None
    """unique id of the stream schedule. Might be absent for live streams"""
    user_id: str
    """unique id of the user hosting this stream"""
    cast_id: Optional[str]
    """unique id of the stream cast. Might be absent"""
    username: str
    """channel name"""
    name: str
    """visible name of the user"""
    avatar_url: str
    """link to the user's avatar"""
    playlist_url: Optional[str] = None
    """link to the underlying hls playlist of the livestream"""

    is_live: bool
    """indicates that livestream is currently live"""
    price: int
    """access price, in points"""

    def __str__(self):
        last_line = ''
        scheduled = self.scheduled
        if scheduled:
            last_line = '\nscheduled to {}'.format(scheduled.strftime('%Y-%m-%d %H:%M'))
        elif self.is_live:
            last_line = '\n[Live]'
        if self.price > 0:
            last_line += f' [{self.price}pt]'
        template = '{}\n{}\npublished by {}'
        return template.format(self.url, self.title, self.name) + last_line

    def __repr__(self):
        template = '{:<8} [{}] {}'
        return template.format(self.name, self.scheduled or self.start, self.title[:60])

    def get_uid(self) -> str:
        return f'{self.stream_id}'

    def as_embed(self) -> dict:
        embed: Dict[str, Any] = {
            'title': self.title,
            # 'description': self.description,
            'url': self.url,
            'color': None,
            'author': {'name': self.name, 'url': self.url, 'icon_url': self.avatar_url},
            'image': {'url': self.thumbnail_url},
            'fields': []
        }
        footer = ''
        if self.scheduled is not None:
            scheduled = self.scheduled.strftime('%Y-%m-%d %H:%M')
            embed['fields'].append({'name': 'Scheduled:', 'value': scheduled, 'inline': True})
        if self.is_live:
            embed['fields'].append({'name': '[Live]', 'value': '', 'inline': True})
        if self.price > 0:
            embed['fields'].append({'name': f'[{self.price}pt]', 'value': '', 'inline': True})
        embed['footer'] = {'text': footer}
        return embed


@dataclasses.dataclass
class Cast:
    username: str
    name: str
    avatar_url: str
    user_id: str
    cast_id: Optional[str]

    @classmethod
    def from_cast_data(cls, data: dict) -> 'Cast':
        try:
            info = data['agencySecret']
            return cls(
                username=info['username'],
                name=info['name'],
                avatar_url=data['profileImageUrl'],
                user_id=data['uuid'],
                cast_id=info['uuid']
            )
        except (TypeError, KeyError, ValueError) as e:
            raise ValueError(f'failed to parse cast info: {type(e)} {e}') from e

    @classmethod
    def from_owner_data(cls, data: dict) -> 'Cast':
        try:
            return cls(
                username=data['username'],
                name=data['name'],
                avatar_url=data['profileImageUrl'],
                user_id=data['uuid'],
                cast_id=None
            )
        except (TypeError, KeyError, ValueError) as e:
            raise ValueError(f'failed to parse owner info: {type(e)} {e}') from e

def parse_live_record(data: dict, cast: Optional[Cast] = None) -> WithnyRecord:
    if cast is None:
        if 'cast' in data:
            cast = Cast.from_cast_data(data['cast'])
        elif 'owner' in data:
            cast = Cast.from_owner_data(data['owner'])
        else:
            raise ValueError(f'no user info in data')
    started_at = parse_date_fields(data, ['startedAt', 'actualStartedAt'])
    ended_at = parse_date_fields(data, ['closedAt', 'actualClosedAt'])
    return WithnyRecord(
        url=f'https://www.withny.fun/channels/{cast.username}',
        stream_id=data['uuid'],
        title=data['title'],
        description=data['about'],
        thumbnail_url=data['thumbnailUrl'],
        start=started_at,
        end=ended_at,
        scheduled=None,
        user_id=cast.user_id,
        cast_id=cast.cast_id,
        username=cast.username,
        name=cast.name,
        avatar_url=cast.avatar_url,
        is_live=started_at is not None and ended_at is None,
        price=data['price']
    )


def parse_schedule_record(data: dict) -> WithnyRecord:
    cast = Cast.from_cast_data(data['cast'])
    scheduled = parse_date_fields(data, ['startAt', 'scheduledStartedAt'])
    schedule_id = data['uuid']

    stream_record = parse_live_record(data['stream'], cast)
    stream_record.scheduled = scheduled
    stream_record.schedule_id = schedule_id

    return stream_record


def format_timestamp(ts: datetime.datetime) -> str:
    ts = ts.astimezone(datetime.timezone.utc)
    text = ts.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
    return text


def utcnow_with_offset(days: int = 0, hours: int = 0) -> datetime.datetime:
    return utcnow() + datetime.timedelta(days=days, hours=hours)


def parse_date_fields(data: Dict[str, Any], fields: List[str]) -> Optional[datetime.datetime]:
    """try parsing given fields of the data as datetime, return first successfully parsed or None"""
    for field in fields:
        field_text = data.get(field)
        if field_text is not None:
            return dateutil.parser.parse(field_text)
    return None



import asyncio
import json
import logging
import pathlib
import urllib.parse
from collections import defaultdict
from typing import Callable, Dict, List, Optional

import dateutil.zoneinfo
from aiohttp import web
from pydantic import BaseModel, ValidationError

from avtdl.core import formatters, info
from avtdl.core.actors import Actor
from avtdl.core.cache import FileCache
from avtdl.core.chain import Chain
from avtdl.core.config import ConfigParser, ConfigurationError, SettingsSection
from avtdl.core.info import get_known_plugins, get_plugin_type, render_markdown
from avtdl.core.interfaces import AbstractRecordsStorage, Record
from avtdl.core.plugins import Plugins
from avtdl.core.runtime import RuntimeContext, TaskStatus, TerminatedAction
from avtdl.core.utils import JSONType, strip_text, write_file
from avtdl.core.yaml import merge_data, yaml_dump

RECORDS_PER_PAGE = 32


def serialize_config(settings: SettingsSection,
                     actors: Dict[str, Actor],
                     chains: Dict[str, Chain]) -> str:
    config = ConfigParser.serialize(settings, actors, chains)
    conf = config.model_dump_json()
    return conf


def json_dumps(obj):
    def default(o):
        if isinstance(o, pathlib.Path):
            return str(o)
        raise TypeError(f'Object of type {o.__class__.__name__} is not JSON serializable')

    return json.dumps(obj, default=default)


def get_actor_description(name: str) -> str:
    plugin, config, entity = Plugins.get_actor_factories(name)
    md_description = info.render_doc(plugin)
    html_description = info.render_markdown(md_description)
    return html_description


def get_schema(model: BaseModel) -> dict:
    schema = model.model_json_schema(mode='serialization')
    render_descriptions(schema)
    return schema


def render_descriptions(schema) -> None:
    if isinstance(schema, dict):
        if 'description' in schema:
            text = schema['description']
            if isinstance(text, str) and len(text) > 0:
                end = '' if text[-1] in '.?!,' else '.'
                text = text[0].upper() + text[1:] + end
            schema['description'] = info.render_markdown(text)
        else:
            for subschema in schema.values():
                render_descriptions(subschema)
    elif isinstance(schema, list) or isinstance(schema, tuple):
        for subschema in schema:
            render_descriptions(subschema)
    else:
        return


def get_conf_schema(actor_name: str) -> dict:
    plugin, config, entity = Plugins.get_actor_factories(actor_name)
    return get_schema(config)


def get_entity_schema(actor_name: str) -> dict:
    plugin, config, entity = Plugins.get_actor_factories(actor_name)
    return get_schema(entity)


def record_preview(record: Record, representation: str = 'text') -> str:
    if representation == 'text':
        return str(record).replace('\n', '<br>\n')
    elif representation == 'json':
        return record.as_json(indent=4)
    elif representation == 'short':
        return repr(record)
    else:
        return str(record)


class ActorModel(BaseModel):
    type: Optional[str]
    description: str
    config_schema: dict
    entity_schema: dict


class WebUI:
    WEBROOT: pathlib.Path = pathlib.Path(__file__).parent.parent.resolve() / 'ui'
    CACHE_ROUTE = '/cache'
    RESTART_DELAY: int = 3

    def __init__(self, config_path: pathlib.Path, config, ctx: RuntimeContext, settings: SettingsSection,
                 actors: Dict[str, Actor], chains: Dict[str, Chain]):
        self.logger = logging.getLogger('webui')
        self.host = settings.host
        self.port = settings.port
        self.config_path = config_path
        self.config_base = config
        self.ctx = ctx
        self.settings = settings
        self.actors = actors
        self.chains = chains
        self.restart_pending = False
        self.cache = FileCache(self.settings.cache_directory, '.part')
        self._actors_models = self.generate_actors_models()
        self.routes: List[web.AbstractRouteDef] = []

        self.routes.append(web.get('/favicon.ico', self.favicon))
        self.routes.append(web.get('/chains', self.show_chains))
        self.routes.append(web.get('/actors', self.actors_models))
        self.routes.append(web.get('/settings', self.settings_schema))
        self.routes.append(web.get('/config', self.show_config))
        self.routes.append(web.post('/config', self.store_config))
        self.routes.append(web.get('/timezones', self.timezones))
        self.routes.append(web.get('/motd', self.motd))
        self.routes.append(web.get('/history', self.history))
        self.routes.append(web.get('/tasks', self.tasks))

        self.routes.append(web.get('/ui/info/info.html', self.info_webui))

        self.routes.append(web.get('/viewable', self.viewable_plugins))
        self.routes.append(web.get('/records', self.records))
        self.routes.append(web.static(self.CACHE_ROUTE, self.cache.cache_directory))

        self.routes.append(web.get('/', self.index))
        self.routes.append(web.static('/ui', self.WEBROOT, append_version=True))
        self.routes.append(web.static('/res', self.WEBROOT / 'misc', append_version=True))

    async def favicon(self, _: web.Request) -> web.FileResponse:
        return web.FileResponse(self.WEBROOT / 'misc/favicon.svg')

    async def index(self, request):
        raise web.HTTPFound('/ui/conf/config.html')

    async def timezones(self, _: web.Request) -> web.Response:
        zones = list(dateutil.zoneinfo.get_zonefile_instance().zones.keys())
        return web.json_response(zones)

    async def show_chains(self, _: web.Request) -> web.Response:
        data = {name: chain.conf.model_dump() for name, chain in self.chains.items()}
        return web.json_response(data)

    @staticmethod
    def generate_actors_models() -> Dict[str, dict]:
        data: Dict[str, dict] = {}

        for name in get_known_plugins():
            data[name] = ActorModel(
                type=get_plugin_type(name),
                description=get_actor_description(name),
                config_schema=get_conf_schema(name),
                entity_schema=get_entity_schema(name)
            ).model_dump()
        return data

    async def actors_models(self, _: web.Request) -> web.Response:
        return web.json_response(self._actors_models, dumps=json_dumps)

    async def settings_schema(self, _: web.Request) -> web.Response:
        schema = self.settings.model_json_schema(mode='serialization')
        render_descriptions(schema)
        return web.json_response(schema, dumps=json_dumps)

    async def show_config(self, _: web.Request):
        data = serialize_config(self.settings, self.actors, self.chains)
        return web.Response(text=data)

    async def store_config(self, request: web.Request) -> web.Response:
        mode = request.query.get('mode', 'check')
        if not mode in ['check', 'store', 'reload']:
            raise web.HTTPBadRequest(text=f'unexpected "mode" parameter {mode}')
        try:
            conf = await request.json()
            parsed_config = ConfigParser.validate(conf)
            config_encoding = parsed_config.settings.encoding

            updated_config = merge_data(self.config_base, conf)
            raw_yaml = yaml_dump(updated_config)
            if mode == 'check':
                return web.Response(text='Config has been validated successfully')

            self.config_base = updated_config

            try:
                write_file(self.config_path, raw_yaml, encoding=config_encoding, backups=10)
            except Exception as e:
                raise web.HTTPInternalServerError(
                    text=f'failed to store config in "{self.config_path}": {e or type(e)}')
            if mode == 'reload':
                self.restart_pending = True
                self.ctx.controller.terminate_after(self.RESTART_DELAY, TerminatedAction.RESTART)
                return web.Response(
                    text=f'Updated config successfully stored in "{self.config_path}". Restarting in a few seconds (the page will reload).')
            else:
                text = f'Updated config successfully stored in "{self.config_path}". It will be used after next restart.'
                return web.Response(text=text)
        except ConfigurationError as e:
            if e.__cause__ is None:
                raise web.HTTPBadRequest(text=f'Malformed configuration error {type(e)}: {e}')
            if not isinstance(e.__cause__, ValidationError):
                raise web.HTTPBadRequest(text=f'Configuration error is {type(e)}, expected ValidationError: {e}')
            data = e.__cause__.errors()
            for error in data:
                if 'url' in error:
                    error.pop('url', None)  # type: ignore
                if 'ctx' in error:
                    error.pop('ctx')
                if 'msg' in error:
                    error['msg'] = strip_text(error['msg'], 'Value error, ')

            return web.json_response(data=data, dumps=json_dumps, status=422, reason='Bad config')
        except web.HTTPError:
            raise
        except Exception as e:
            raise web.HTTPBadRequest(text=f'{type(e)}: {e}')

    async def motd(self, _: web.Request) -> web.Response:
        if self.restart_pending:
            raise web.HTTPServiceUnavailable(headers={'Retry-After': str(self.RESTART_DELAY)})
        motd = f'''
Server is up and running, working directory is "{pathlib.Path('.').resolve()}".
Configuration contains {len(self.actors)} actors and {len(self.chains)} chains, loaded from "{self.config_path.resolve()}".
'''
        data = {'motd': motd}
        return web.json_response(data, dumps=json_dumps)

    async def info_webui(self, _: web.Request) -> web.Response:

        template_path = self.WEBROOT / 'info/info.html'
        template = template_path.read_text(encoding='utf8')

        document_path = self.WEBROOT / 'info/info.md'
        document = document_path.read_text(encoding='utf8')
        body = render_markdown(document)

        html = template.replace('{{body}}', body)
        return web.Response(text=html, content_type='text/html')

    async def history(self, request: web.Request) -> web.Response:
        actor = request.query.get('actor')
        entity = request.query.get('entity')
        chain = request.query.get('chain', '')
        representation = request.query.get('repr', 'text')
        if actor is None or entity is None:
            raise web.HTTPBadRequest(
                text=f'not enough arguments. Got actor="{actor}", entity="{entity}", chain="{chain}"')
        incoming = self.ctx.bus.get_history(actor, entity, chain, 'in')
        outgoing = self.ctx.bus.get_history(actor, entity, chain, 'out')
        data_structure = [
            (f'Incoming records (most recent)', incoming),
            (f'Outgoing records (most recent)', outgoing)
        ]
        data = {}

        for title, content in data_structure:
            rows = []
            for record in content:
                row = [int(record.created_at.timestamp() * 1000),
                       record.origin,
                       record.chain,
                       record_preview(record, representation)]
                rows.append(row)
            data[title] = rows
        return web.json_response(data, dumps=json_dumps)

    @staticmethod
    def render_status_data(status_list: List[TaskStatus], include_empty: bool = False) -> dict:
        if not status_list:
            return {}
        headers = ['Actor', 'Entity', 'Info', 'Record']
        unsorted_data: dict = defaultdict(lambda: {'headers': headers, 'rows': []})
        for status in status_list:
            if not include_empty and status.is_empty():
                continue
            record = record_preview(status.record) if status.record else ''
            actor_type = get_plugin_type(status.actor or 'other') or 'Other'
            row = [status.actor, status.entity, status.status, record]
            unsorted_data[actor_type]['rows'].append(row)
        sorted_data = sorted(((k, v) for k, v in unsorted_data.items()), key=lambda x: x[0])
        data: dict = defaultdict(lambda: {'headers': headers, 'rows': []})
        data.update({k: v for k, v in sorted_data})
        return data

    async def tasks(self, request: web.Request) -> web.Response:
        show_empty = request.query.get('empty') is not None
        actor_name = request.query.get('actor')
        if actor_name is not None:
            actor = self.actors.get(actor_name) if actor_name is not None else None
            if actor is None:
                raise web.HTTPBadRequest(text=f'actor "{actor_name}" is not found')
            controller = actor.ctx.controller
        else:
            controller = self.ctx.controller

        status_list = controller.get_status()
        data = self.render_status_data(status_list, show_empty)
        return web.json_response(data, dumps=json_dumps)

    async def viewable_plugins(self, _: web.Request) -> web.Response:
        internal: Dict[str, Dict[str, Dict[str, Dict[str, str]]]] = defaultdict(dict)
        viewable: Dict[str, Dict[str, str]] = defaultdict(dict)
        for actor_name, actor in self.actors.items():
            entities: dict = defaultdict(dict)
            for view_name in [None, *actor.entities.keys()]:
                db = actor.get_records_storage(view_name)
                if db is None:
                    continue
                feeds = db.feeds()
                if not feeds:
                    continue
                queries: Dict[str, str] = {}
                if len(feeds) > 1:
                    params = {'actor': actor_name,}
                    if view_name:
                        params['view'] = view_name
                    queries['[Everything]'] = urllib.parse.urlencode(params)
                for entity_name, record_count in feeds:
                    params = {'actor': actor_name, 'entity': entity_name}
                    if view_name:
                        params['view'] = view_name
                    query = urllib.parse.urlencode(params)
                    queries[f'{entity_name} ({record_count})'] = query
                if view_name is not None:
                    entities[view_name] = queries
                else:
                    entities = queries
            actor_type = get_plugin_type(actor_name) or 'Other'
            if actor.conf.name == 'view':
                viewable = entities
            elif actor.get_records_storage() is not None:
                internal[actor_type][actor_name] = entities
        data: dict = {}
        if viewable:
            data['View stored records'] = viewable
        if internal:
            data['Internal databases'] = internal
        return web.json_response(data, dumps=json_dumps)

    def _get_embed_image_rewriter(self, record: Record) -> Callable[[str], Optional[str]]:
        def rewriter(image_url: str) -> Optional[str]:
            image_file = self.cache.retrieve(record, image_url)
            if image_file is None:
                return None
            try:
                relative = image_file.relative_to(self.cache.cache_directory)
            except Exception:
                msg = f'cached file "{image_file}" for url "{image_url}" is not relative to "{self.cache.cache_directory}"'
                self.logger.warning(msg)
                return None
            resource = str(self.CACHE_ROUTE / relative)
            return resource
        return rewriter

    def render_record(self, record: Record) -> JSONType:
        embeds = formatters.MessageFormatter.make_embeds(record, False)
        rewriter = self._get_embed_image_rewriter(record)
        for embed in embeds:
            formatters.MessageFormatter.rewrite_embed_links(embed, rewriter)
        message = formatters.MessageFormatter.make_message(embeds)
        return message

    async def records(self, request: web.Request) -> web.Response:
        actor_name = request.query.get('actor')
        entity_name = request.query.get('entity') or None
        view_name = request.query.get('view') or None
        page_num = request.query.get('page')
        page_size = request.query.get('size', RECORDS_PER_PAGE)

        if page_num is not None:
            try:
                page = int(page_num)
            except ValueError:
                raise web.HTTPBadRequest(text=f'page number must be positive integer')
        else:
            page = None
        try:
            per_page = int(page_size)
        except ValueError:
            raise web.HTTPBadRequest(text=f'page size must be positive integer')

        if actor_name is None:
            raise web.HTTPBadRequest(text=f'missing "actor" parameter')
        if actor_name not in self.actors:
            raise web.HTTPBadRequest(text=f'actor {actor_name} not found')

        actor = self.actors[actor_name]

        db: Optional[AbstractRecordsStorage] = actor.get_records_storage(view_name)
        if db is None:
            raise web.HTTPBadRequest(text=f'actor {actor_name} does not have persistent storage')
        records = db.load_page(page, per_page, feed=entity_name)
        total_pages = db.page_count(per_page, entity_name)
        records_view = [self.render_record(record) for record in records]

        data = {
            'total': total_pages,
            'current': page or total_pages,
            'feed': entity_name,
            'records': records_view
        }
        return web.json_response(data, dumps=json_dumps)


async def run_app(webui: WebUI):
    app = web.Application()
    app.add_routes(webui.routes)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, webui.host, webui.port)
    webui.logger.debug('starting server...')
    try:
        await site.start()
    except Exception as e:
        webui.logger.exception(f'failed to start server: {e}')
        return
    webui.logger.info(f'server is running on http://{webui.host}:{webui.port}')
    try:
        await asyncio.Future()
    except (KeyboardInterrupt, asyncio.CancelledError):
        webui.logger.info('stopping server...')
        await runner.cleanup()
        webui.logger.debug('server stopped')


async def run(config_path: pathlib.Path, config, ctx: RuntimeContext, settings: SettingsSection,
              actors: Dict[str, Actor], chains: Dict[str, Chain]):
    webui = WebUI(config_path, config, ctx, settings, actors, chains)
    await run_app(webui)



import datetime
import json
import logging
import re
import textwrap
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Type, Union

import markdown
from markdown.extensions.toc import TocExtension
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

from avtdl.core.actors import Action, Filter, Monitor
from avtdl.core.plugins import Plugins
from avtdl.core.yaml import yaml_load

# plugins names starting with this are excluded from docs
INTERNAL_PLUGINS_PATTERN = 'utils'

HELP_FILE_STATIC_PART = f'''
<!-- This file is autogenerated as part of avtdl at {datetime.datetime.utcnow()} -->

## Description and configuration of available plugins

This page provides details on format and possible values of configuration
options of the plugins and plugins' entities, as well as description of
records produced by them.

High-level overview of the configuration file structure can be found in
[README](README.md#configuration-file-format). Example configurations
for common workflows are available [here](EXAMPLES.md).

In each plugin section, *Plugin configuration options*, if present, describes
plugin-wide settings, placed under the `config` part inside the plugin config.
The *Entity configuration options* lists settings for each entity inside the
`entities` list.

A few common options used by multiple plugins, most notable the templating rules,
are described in greater details in the [Common options](README.md#common-options)
section of the README.

### Table of content:
<!-- [TOC] -->

{{TOC}}

<!-- [TOC] -->
---
'''

STRUCTURE = '''
## Monitors:

{}

## Filters:

{}

## Actions:

{}
'''

# implicitly relies on first of the lines inside `description` being a short title
PLUGIN_INFO_TEMPLATE = '''
### `{name}` - {description}
'''

PLUGIN_OPTIONS_TEMPLATE = '''
#### Plugin configuration options:
{config}

'''
ENTITY_OPTIONS_TEMPLATE = '''
#### Entity configuration options:
{entity}
'''

LIST_ITEM_TEMPLATE = '* `{name}`: {description}'

ASSOCIATED_RECORDS_TEMPLATE = '''
#### Produced records types:
'''

COLLAPSIBLE_ITEM_TEMPLATE = '''
<details markdown="block">
  <summary>{title}</summary>

{content}

</details>
'''


def get_plugin_info(plugin_name: str) -> str:
    plugin, config, entity = Plugins.get_actor_factories(plugin_name)
    description = render_doc(plugin)
    text = [PLUGIN_INFO_TEMPLATE.format(name=plugin_name, description=description)]
    config_info = get_config_model_info(config)
    if config_info:
        text.append(PLUGIN_OPTIONS_TEMPLATE.format(config=config_info))
    entity_info = get_entity_model_info(entity)
    if entity_info:
        text.append(ENTITY_OPTIONS_TEMPLATE.format(entity=entity_info))
    associated_records = Plugins.get_associated_records(plugin_name)
    if associated_records:
        records_text = [ASSOCIATED_RECORDS_TEMPLATE]
        for record_type in associated_records:
            record_info = get_record_model_info(record_type)
            record_text = COLLAPSIBLE_ITEM_TEMPLATE.format(title=record_type.__name__, content=record_info)
            records_text.append(record_text)
        text.extend(records_text)
    return '\n'.join(text)


def get_model_info(model: Type[BaseModel], skip_name: bool = False) -> str:
    info: List[str] = []
    description = render_doc(model)
    if description:
        info.append(description)
        info.append('\n')  # ensure newline before list to make it render correctly
    required_fields = []
    not_required_fields = []
    for name, field_info in model.model_fields.items():
        if skip_name and name == 'name':
            continue
        if field_info.exclude:
            continue
        field_description = render_field_info(field_info, skip_details=False)
        field_description_text = LIST_ITEM_TEMPLATE.format(name=name, description=field_description)
        if has_default(field_info):
            not_required_fields.append(field_description_text)
        else:
            required_fields.append(field_description_text)
    info.extend(required_fields)
    if required_fields and not_required_fields:
        info.append('##### ')  # separate mandatory and non-mandatory fields in two lists
    info.extend(not_required_fields)
    return '\n'.join(info)


def get_config_model_info(model: Type[BaseModel]) -> str:
    return get_model_info(model, skip_name=True)


def get_entity_model_info(model: Type[BaseModel]) -> str:
    return get_model_info(model, skip_name=False)


def get_record_model_info(model: Type[BaseModel]) -> str:
    info: List[str] = []
    description = render_doc(model)
    if description:
        info.append(description)
        info.append('\n')  # ensure newline before list to make it render correctly
    for name, field_info in model.model_fields.items():
        if field_info.exclude:
            continue
        field_description = render_field_info(field_info, skip_details=True)
        field_description_text = LIST_ITEM_TEMPLATE.format(name=name, description=field_description)
        info.append(field_description_text)
    return '\n'.join(info)


def render_doc(model: Type[BaseModel]) -> str:
    if model.__doc__:
        text = textwrap.dedent(model.__doc__).strip('\n')
        return text
    return ''


def render_field_info(field_info: FieldInfo, skip_details=False) -> str:
    FIELD_INFO_TEMPLATE = '{description}. {details}'
    default = get_default(field_info)
    if skip_details:
        details = ''
    elif default:
        details = f'Default value is `{default}`.'
    else:
        details = 'Required.' if field_info.is_required() else 'Not required.'
    description = field_info.description or ''
    return FIELD_INFO_TEMPLATE.format(details=details, description=description)


def has_default(field_info: FieldInfo) -> bool:
    return field_info.default is not PydanticUndefined


def get_default(field_info: FieldInfo) -> Optional[str]:
    """Return text describing default value of given FieldInfo if set"""
    value = field_info.default
    if value is PydanticUndefined:
        return None
    if value is None:
        return None
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, Enum):
        return value.value
    if isinstance(value, str):
        if not value.strip(' \t\r\n'):
            return None
        return value
    if isinstance(value, dict):
        return ', '.join(f'"{k}": "{v}"' for k, v in value.items())
    if isinstance(value, type):
        return value.__name__
    return str(value)


def get_plugin_type(name: str) -> Optional[str]:
    plugin = Plugins.known[Plugins.kind.ACTOR].get(name)
    if plugin is None:
        return None
    if name.startswith(INTERNAL_PLUGINS_PATTERN):
        return 'Utils'
    if issubclass(plugin, Monitor):
        return 'Monitors'
    if issubclass(plugin, Filter):
        return 'Filters'
    if issubclass(plugin, Action):
        return 'Actions'
    return None


def get_plugins_descriptions(plugin_type: type = object):
    descriptions = {}
    for name, constructor in Plugins.known[Plugins.kind.ACTOR].items():
        if name.startswith(INTERNAL_PLUGINS_PATTERN):
            continue
        if not issubclass(constructor, plugin_type):
            continue
        descriptions[name] = get_plugin_info(name)
    return descriptions


def render_plugins_descriptions() -> str:
    """load available plugins and generate a help file in markdown from docstrings"""
    SEPARATOR = '\n---\n'

    Plugins.load()
    monitors = get_plugins_descriptions(Monitor)
    filters = get_plugins_descriptions(Filter)
    actions = get_plugins_descriptions(Action)
    descriptions = [SEPARATOR.join(section.values()) for section in [monitors, filters, actions]]
    text = STRUCTURE.format(*descriptions)
    return text


md = markdown.Markdown(extensions=[TocExtension(toc_depth=3), 'md_in_html', 'fenced_code'])


def render_markdown(text: str) -> str:
    """convert markdown to html fragment"""
    html = md.convert(text)
    return html


def generate_plugins_description():
    text = render_plugins_descriptions()
    full_text = HELP_FILE_STATIC_PART + text
    full_text = TOC.insert_toc(text, full_text)
    return full_text


def generate_version_string() -> str:
    try:
        from avtdl._version import __version__
    except ModuleNotFoundError:
        __version__ = '[unknown version]'
    Plugins.load()
    version = f'avtdl {__version__} with plugins:'
    known_plugins = ', '.join([name for name in Plugins.known[Plugins.kind.ACTOR].keys()])
    known_plugins = textwrap.fill(known_plugins, initial_indent='    ', subsequent_indent='    ')
    text = version + '\n' + known_plugins
    return text


class TOCToken(BaseModel):
    name: str
    id: str
    level: int
    children: List['TOCToken']


class TOC:

    @staticmethod
    def slugify(text: str, separator='-') -> str:
        text = re.sub(r'[^\w\s_-]', '', text).strip().lower()
        text = re.sub(r'[\s]+', separator, text)
        return text

    @classmethod
    def generate_toc_tokens(cls, text: str, toc_depth: Union[int, str]) -> List[TOCToken]:
        md = markdown.Markdown(output_format='html',
                               extensions=['md_in_html', TocExtension(toc_depth=toc_depth, slugify=cls.slugify)])
        _ = md.convert(text)
        tokens = [TOCToken.model_validate(token) for token in md.toc_tokens]  # type: ignore
        return tokens

    @staticmethod
    def generate_toc_line(t: TOCToken, level: int = 0, indent: str = '  ') -> str:
        line = f'{level * indent}* [{t.name}](#{t.id})'
        return line

    @classmethod
    def generate_toc_md(cls, tokens: List[TOCToken], level: int = 0, indent: str = '  ') -> List[str]:
        md: List[str] = []
        for token in tokens:
            line = cls.generate_toc_line(token, level, indent)
            md.append(line)
            children_md = cls.generate_toc_md(token.children, level + 1, indent)
            md.extend(children_md)
        return md

    @classmethod
    def generate_toc(cls, text: str, toc_depth: Union[int, str], indent: str = '  ') -> str:
        tokens = cls.generate_toc_tokens(text, toc_depth)
        md_lines = cls.generate_toc_md(tokens, indent=indent)
        md = '\n'.join(md_lines)
        return md

    @classmethod
    def insert_toc(cls, source_text: str, target_text: str,
                   toc_marker: str = '{TOC}', toc_depth='2-3', indent: str = '  ') -> str:
        toc = cls.generate_toc(source_text, toc_depth, indent)
        result = target_text.replace(toc_marker, toc)
        return result


def get_known_plugins() -> List[str]:
    plugins_by_type: Dict[str, List[str]] = {
        'Monitors': [],
        'Filters': [],
        'Actions': [],
        'Other': []
    }
    for name, plugin in Plugins.known[Plugins.kind.ACTOR].items():
        plugin_type = get_plugin_type(name)
        if plugin_type not in plugins_by_type:
            plugin_type = 'Other'
        plugins_by_type[plugin_type].append(name)
    plugins = []
    for plugin_list in plugins_by_type.values():
        plugins.extend(plugin_list)
    return plugins


@dataclass
class ConfigExample:
    title: str
    """name of the example"""
    description: str
    """brief description and comments, rendered as html"""
    text_yaml: str
    """example as yaml"""
    text_json: str
    """example as json"""

    @classmethod
    def make(cls, title, description, example) -> 'ConfigExample':
        rendered_description = render_markdown(description)
        config = yaml_load(example)
        text_json = json.dumps(config, indent=4, ensure_ascii=False)
        return cls(title=title,
                   description=rendered_description,
                   text_yaml=example,
                   text_json=text_json
        )


def load_examples(file: Path = Path(__file__).parent.parent.parent / 'EXAMPLES.md') -> List[ConfigExample]:
    """
    Load and parse example configurations from EXAMPLES.md
    """
    try:
        text = file.read_text(encoding='utf8')
    except Exception as e:
        logger = logging.getLogger()
        logger.warning(f'[load_examples] failed to load examples from "{file}": {e}')
        return []

    pattern = re.compile(r'^#{1,6} ([^\n]*)$\n\n([^#]*?)^```yaml\n?(.*?)\n?```', re.DOTALL | re.MULTILINE)
    results = re.findall(pattern, text)
    examples = [ConfigExample.make(*result) for result in results]
    return examples



import dataclasses
import datetime
import logging
import re
from email.utils import mktime_tz
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import lxml.html

from avtdl.core.interfaces import Record
from avtdl.core.utils import is_url


def sanitize_filename(name: str, collapse: bool = False) -> str:
    """Replace symbols not allowed in file names on NTFS with underscores"""
    pattern = r'[\\/:*?"<>|]+' if collapse else r'[\\/:*?"<>|]'
    return re.sub(pattern, "_", name)


def make_datetime(items) -> datetime.datetime:
    """take 10-tuple and return datetime object with UTC timezone"""
    if len(items) == 9:
        items = *items, None
    if len(items) != 10:
        raise ValueError(f'Expected tuple with 10 elements, got {len(items)}')
    timestamp = mktime_tz(items)
    return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)


def html_from_string(html: str, base_url: Optional[str] = None) -> lxml.html.HtmlElement:
    try:
        root: lxml.html.HtmlElement = lxml.html.fromstring(html)
        if base_url is not None:
            root.make_links_absolute(base_url=base_url, handle_failures='ignore')
        return root
    except Exception as e:
        logging.getLogger('html_to_text').exception(e)
        raise


def html_to_text(html: str, base_url: Optional[str] = None, markdown: bool = False, strip_img: bool = False) -> str:
    """Take html fragment, try to parse it and convert to text using lxml
    Convert links to markdown representation if markdown is True"""
    try:
        root = html_from_string(html, base_url)
    except Exception:
        return html
    text_nodes = html_to_text2(root, Context(plaintext = not markdown, strip_img=strip_img))
    return ''.join(text_nodes)


@dataclasses.dataclass
class Context:
    plaintext: bool = False # element should be rendered in plaintext mode
    strip_img: bool = False # all <img> should be dropped
    a: bool = False # inside <a>
    pre: bool = False # inside <pre>
    code: bool = False # inside <code>


def html_to_text2(elem: lxml.html.HtmlElement, ctx: Context) -> List[str]:
    before = None
    after = None
    text = elem.text
    tail = elem.tail
    children = None

    if elem.tag == 'pre' or elem.tag == 'code':
        if not ctx.plaintext and not ctx.code and not ctx.pre:
            ctx = dataclasses.replace(ctx, **{elem.tag: True})
            children = children_to_text2(elem, ctx)
            if elem.tag == 'pre' or is_multiline(elem, children):
                before = after = '```'
            else:
                before = after = '`'
        ctx = dataclasses.replace(ctx, **{elem.tag: True})
    if elem.tag == 'p':
        after = '\n\n'
    if elem.tag == 'br':
        after = '\n'
    if elem.tag == 'a':
        ctx = dataclasses.replace(ctx, a=True)
        href = elem.get('href')
        if href is not None:
            children = children_to_text2(elem, ctx)

            text_is_empty = not elem.text and (not children or not any(children))
            if text_is_empty: # link has no content, likely because it contained now stripped image
                children = [href]

            if ctx.plaintext:
                after = f' ({href})'
            else:
                before, after = '[', f']({href})'
    if elem.tag == 'img':
        if not ctx.strip_img:
            src = elem.get('src')
            if src is not None:
                if ctx.plaintext:
                    after = f'\n{src}\n'
                else:
                    # render images as regular links, they should already be included in attachments
                    # for regular links containing image, drop image link and only keep text
                    text = elem.get('alt') or elem.get('title') or src
                    if ctx.a:
                        after = text
                    else:
                        after = f'\n[{text}]({src})\n'

    if children is None:
        children = children_to_text2(elem, ctx)
    if not ctx.plaintext and not (ctx.pre or ctx.code or ctx.a):
        if text is not None:
            text = escape_markdown(text)
        # if tail is not None:
        #     tail = escape_markdown(tail)
    if not ctx.pre:
        if text is not None:
            text = re.sub(r'\n+$', '', text)
        if tail is not None:
            tail = re.sub(r'\n+$', '', tail)

    nodes = [before, text, *children, after, tail]
    real_nodes = [node for node in nodes if node is not None]

    return real_nodes


def is_multiline(elem: lxml.html.HtmlElement, children: List[str]) -> bool:
    if elem.text is not None and '\n' in elem.text:
        return True
    for text_node in children:
        if '\n' in text_node:
            return True
    return False


def children_to_text2(elem: lxml.html.HtmlElement, ctx: Context) -> List[str]:
    nodes = []
    for child in elem.iterchildren():
        child_nodes = html_to_text2(child, ctx=ctx)
        nodes.extend(child_nodes)
    return nodes


def escape_markdown(text: str) -> str:
    """Escape markdown special characters"""
    escaped = re.sub(r'([\\`*_{}\[\]()#+-.!])', r'\\\1', text)
    return escaped


def html_images(html: str, base_url: Optional[str]) -> List[str]:
    """take html fragment, try to parse it and extract image links"""
    try:
        root = html_from_string(html, base_url)
    except Exception:
        return []
    images = [elem.get('src') for elem in root.iter() if elem.tag == 'img' and elem.get('src')]
    return images


class OutputFormat(str, Enum):
    text = 'text'
    repr = 'short'
    json = 'json'
    pretty_json = 'pretty_json'
    hash = 'hash'


class Fmt:
    """Helper class to interpolate format string from config using data from Record"""

    @classmethod
    def format(cls, fmt: str, record: Record, missing: Optional[str] = None, tz: Optional[datetime.tzinfo] = None,
               sanitize: bool = False, extra: Optional[Dict[str, Any]] = None) -> str:
        """Take string with placeholders like {field} and replace them with record fields"""
        logger = logging.getLogger().getChild('format')
        result = cls.strftime(fmt, datetime.datetime.now(tz))
        record_as_dict = record.model_dump()
        if extra is not None:
            record_as_dict.update(extra)
        placeholders: List[str] = re.findall(r'({[^{}\\]+})', fmt)
        for placeholder in placeholders:
            field = placeholder.strip('{}')
            value = record_as_dict.get(field)
            if value is not None:
                value = cls.format_value(value, sanitize)
                result = result.replace(placeholder, value)
            else:
                if missing is not None:
                    result = result.replace(placeholder, missing)
                else:
                    logger.warning(
                        f'placeholder "{placeholder}" used by format string "{fmt}" is not a field of {record.__class__.__name__} ({record!r}), resulting command is unlikely to be valid')
        result = result.replace(r'\{', '{')
        result = result.replace(r'\}', '}')
        return result

    @classmethod
    def format_value(cls, value: Any, sanitize: bool = False) -> str:
        if value is None:
            value = ''
        elif isinstance(value, datetime.datetime):
            value = cls.date(value)
        else:
            value = str(value)
            if sanitize:
                value = sanitize_filename(value)
        return value

    @classmethod
    def format_path(cls, path: Union[str, Path], record: Record, missing: Optional[str] = None,
                    tz: Optional[datetime.tzinfo] = None, extra: Optional[Dict[str, Any]] = None) -> Path:
        """Take string with placeholders and replace them with record fields, but strip them from bad symbols"""
        fmt = str(path)
        formatted_path = cls.format(fmt, record, missing, tz=tz, sanitize=True, extra=extra)
        return Path(formatted_path)

    @classmethod
    def format_filename(cls, path: Union[str, Path], name: str, record: Record, missing: Optional[str] = None,
                        tz: Optional[datetime.tzinfo] = None, extra: Optional[Dict[str, Any]] = None) -> Path:
        """format file path and filename templates into a Path object"""
        path = cls.format_path(path, record, missing, tz, extra)
        formatted_name = cls.format(name, record, missing, tz=tz, sanitize=True, extra=extra)
        sanitized_name = sanitize_filename(formatted_name)
        return path / sanitized_name

    @classmethod
    def strftime(cls, fmt: str, dt: datetime.datetime) -> str:
        if '%' in fmt:
            fmt = re.sub(r'(%[^aAwdbBmyYHIpMSfzZjUWcxX%GuV])', r'%\1', fmt)
        try:
            return dt.strftime(fmt)
        except ValueError as e:
            logger = logging.getLogger().getChild('format').getChild('strftime')
            logger.debug(f'error adding current date to template "{fmt}": {e}')
            return fmt

    @classmethod
    def date(cls, dt: datetime.datetime) -> str:
        return dt.strftime('%Y-%m-%d %H:%M')

    @classmethod
    def size(cls, size: Union[int, float]) -> str:
        for unit in ['B', 'kB', 'MB', 'GB', 'TB']:
            if size < 1024:
                break
            size /= 1024
        return f"{size:.2f} {unit}"

    @classmethod
    def duration(cls, seconds: int) -> str:
        hours, remainder = divmod(seconds, 3600)
        minutes, seconds = divmod(remainder, 60)
        return f"{hours}:{minutes:02}:{seconds:02}"

    @classmethod
    def dtf(cls, dt: datetime.datetime) -> str:
        """format datetime to Discord timestamp"""
        ts = int(dt.timestamp())
        return f'<t:{ts}>'

    @classmethod
    def save_as(cls, record: Record, output_format: OutputFormat = OutputFormat.text) -> str:
        """Take a record and convert in to string as text/json or sha1"""
        if output_format == OutputFormat.text:
            return str(record)
        if output_format == OutputFormat.repr:
            return repr(record)
        if output_format == OutputFormat.json:
            return record.as_json()
        if output_format == OutputFormat.pretty_json:
            return record.as_json(2)
        if output_format == OutputFormat.hash:
            return record.hash()


class DiscordEmbedLimits:
    TOTAL = 6000
    AUTHOR_NAME = 256
    TITLE = 256
    DESCRIPTION = 4096
    FOOTER_TEXT = 2048
    EMBEDS_PER_MESSAGE = 10


class MessageFormatter:

    @classmethod
    def format(cls, records: List[Record]) -> Tuple[dict, List[Record]]:
        """take records and format them in Discord webhook payload as embeds
        after the limit on embeds is reached, the rest of the records are returned back"""
        embeds: List[Dict[str, Any]] = []
        excess_records = []
        for i, record in enumerate(records):
            record_embeds = cls.make_embeds(record, True)
            if len(embeds) + len(record_embeds) > DiscordEmbedLimits.EMBEDS_PER_MESSAGE:
                excess_records = records[i:]
                break
            else:
                embeds.extend(record_embeds)
        message = cls.make_message(embeds)
        return message, excess_records

    @classmethod
    def make_message(cls, embeds: List[Dict[str, Any]]) -> dict:
        return {
            "content": None,
            "embeds": embeds
        }

    @classmethod
    def make_embeds(cls, record: Record, strip_extra: bool = False) -> List[Dict[str, Any]]:
        embeds = record.as_embed()
        if not isinstance(embeds, list):
            embeds = [embeds]
        if strip_extra:
            for embed in embeds:
                cls.clean_embed(embed)
        else:
            # only adding more extra fields if they are not going to be stripped
            for embed in embeds:
                embed['_timestamp'] = int(record.created_at.timestamp() * 1000)
                embed['_origin'] = record.origin

        return embeds

    @classmethod
    def clean_embed(cls, embed: Dict[str, Any]):
        """remove embed fields starting with underscore"""
        extra_fields = []
        for field in embed:
            if field.startswith('_'):
                extra_fields.append(field)
            elif isinstance(field, dict):
                cls.clean_embed(field)
        for field in extra_fields:
            embed.pop(field)

    @classmethod
    def rewrite_embed_links(cls, embed: Dict[str, Any], rewriter: Callable[[str], Optional[str]]):
        """replace known image urls in the embed with rewriter(url)"""
        image_fields = [('image', 'url'),
                        ('image', 'thumbnail'),
                        ('image', '_preview'),
                        ('author', 'icon_url'),
                        ('footer', 'icon_url')]
        for field_name, subfield_name in image_fields:
            field = embed.get(field_name, None)
            if not isinstance(field, dict):
                continue
            image_url = field.get(subfield_name, None)
            if image_url is not None and is_url(image_url):
                new_url = rewriter(image_url)
                if new_url is not None:
                    embed[field_name][subfield_name] = new_url

    @classmethod
    def check_limits(cls, message: dict) -> bool:
        # doesn't count field.name and field.value number and size in hope it will not change outcome

        total_length = 0
        embeds = message.get('embeds', [])
        if len(embeds) > DiscordEmbedLimits.EMBEDS_PER_MESSAGE:
            return False

        for embed in embeds:
            author_name = len(embed.get('author', {}).get('name', '') or '')
            title = len(embed.get('title') or '')
            description = len(embed.get('description') or '')
            footer_text = len((embed.get('footer') or {}).get('text') or '')
            if author_name > DiscordEmbedLimits.AUTHOR_NAME:
                return False
            if title > DiscordEmbedLimits.TITLE:
                return False
            if description > DiscordEmbedLimits.DESCRIPTION:
                return False
            if footer_text > DiscordEmbedLimits.FOOTER_TEXT:
                return False
            total_length += author_name + title + description + footer_text

        if total_length > DiscordEmbedLimits.TOTAL:
            return False
        return True



import asyncio
import datetime
import json
import logging
import urllib.parse
from http.cookies import SimpleCookie
from typing import Any, Callable, Dict, Optional, Sequence, Tuple

from aiohttp.abc import AbstractCookieJar
from pydantic import Field, FilePath, PositiveInt, SerializeAsAny, field_validator

from avtdl.core.actions import TaskAction, TaskActionConfig, TaskActionEntity
from avtdl.core.db import BaseDbConfig, RecordDB
from avtdl.core.interfaces import Event, EventType, Record
from avtdl.core.plugins import Plugins
from avtdl.core.request import HttpClient, RetrySettings, StateStorage
from avtdl.core.runtime import RuntimeContext, TaskStatus
from avtdl.core.utils import CookieStoreError, JSONType, get_cookie_value, jwt_decode, load_cookies, \
    save_cookies, utcnow
from avtdl.plugins.withny.extractors import WithnyRecord, parse_live_record, parse_schedule_record


@Plugins.register('withny.live', Plugins.kind.ASSOCIATED_RECORD)
class WithnyLiveErrorEvent(Event):
    """Produced on failure to process a livestream"""
    event_type: str = EventType.error
    """text describing the nature of event, can be used to filter classes of events, such as errors"""
    record: SerializeAsAny[Optional[WithnyRecord]] = Field(exclude=True)
    """record that was being processed when this event happened"""

    def __str__(self):
        return f'Processing stream {self.record.stream_id} failed: {self.text}\n[{self.record.name}] {self.record.title}\n{self.record.url}'


def timestamp_now_ms() -> float:
    return datetime.datetime.now().timestamp() * 1000


class AuthToken:
    local_names = {
        'token': 'auth._token.local',
        'token_expiration': 'auth._token_expiration.local',
        'refresh_token': 'auth._refresh_token.local',
        'refresh_token_expiration': 'auth._refresh_token_expiration.local'
    }

    def __init__(self, token: str, token_expiration: str, refresh_token: str, refresh_token_expiration: str):
        self.token = token
        self.token_expiration = token_expiration
        self.refresh_token = refresh_token
        self.refresh_token_expiration = refresh_token_expiration

    def __repr__(self):
        return f'AuthToken(token={self.token}, token_expiration={self.token_expiration}, self.refresh_token={self.refresh_token}, self.refresh_token_expiration={self.refresh_token_expiration})'

    def expired(self) -> bool:
        return int(self.token_expiration) < timestamp_now_ms()

    def refreshable(self) -> bool:
        return int(self.refresh_token_expiration) > timestamp_now_ms()

    @property
    def plain_token(self) -> str:
        return urllib.parse.unquote(self.token)

    def set_cookies(self, jar: AbstractCookieJar):
        cookies: SimpleCookie = SimpleCookie()
        for k, v in self.local_names.items():
            cookies[v] = getattr(self, k)
            cookies[v]['domain'] = 'www.withny.fun'
            cookies[v]['path'] = '/'
        jar.update_cookies(cookies)

    @classmethod
    def from_login_data(cls, data: Dict[str, Any]) -> 'AuthToken':
        token = urllib.parse.quote(f'{data["tokenType"]} {data["token"]}')
        token_content = jwt_decode(data['token'])
        token_expiration = str(token_content['exp'] * 1000)
        refresh_token = data['refreshToken']
        refresh_token_expiration = str(int(token_expiration) + int(datetime.timedelta(days=29).total_seconds() * 1000))
        return cls(
            token,
            token_expiration,
            refresh_token,
            refresh_token_expiration
        )

    @classmethod
    def from_cookies(cls, jar: AbstractCookieJar) -> 'AuthToken':
        data: Dict[str, Any] = {data_key: get_cookie_value(jar, cookie_key) for data_key, cookie_key in cls.local_names.items()}
        for k, v in data.items():
            if v is None:
                raise ValueError(f'{k} value is missing from the jar')
        return cls(**data)


def has_expired(expiration_timestamp: str) -> bool:
    return int(expiration_timestamp) < timestamp_now_ms() + 60000


async def perform_login(client: HttpClient, logger: logging.Logger, username: str, password: str) -> Optional[
    AuthToken]:
    """Make request to log in, update session's cookie jar"""
    url = 'https://www.withny.fun/api/auth/login'
    data = json.dumps({'email': username, 'password': password})
    headers = {'Referer': 'https://www.withny.fun/login', 'Content-Type': 'application/json'}
    return await make_auth_request(client, logger, url, data, headers)


async def refresh_auth(client: HttpClient, logger: logging.Logger) -> Optional[AuthToken]:
    """Make request to refresh auth token, update client's cookie jar"""
    try:
        auth = AuthToken.from_cookies(client.cookie_jar)
    except Exception as e:
        logger.debug(f'[login] failed to refresh token: error when parsing cookies: {type(e)}, {e}')
        return None
    url = 'https://www.withny.fun/api/auth/token'
    data = json.dumps({'refreshToken': auth.refresh_token})
    headers = {'Referer': 'https://www.withny.fun/',
               'Origin': 'https: // www.withny.fun',
               'Content-Type': 'application/json;charset=utf-8',
               'Authorization': auth.plain_token}
    return await make_auth_request(client, logger, url, data, headers)


async def make_auth_request(client: HttpClient, logger: logging.Logger, url: str, data,
                            headers: Optional[dict] = None) -> Optional[AuthToken]:
    result = await client.request_json(url, method='POST', data=data, headers=headers,
                                       settings=RetrySettings(retry_times=1))
    if result is None:
        return None
    try:
        if not isinstance(result, dict):
            raise ValueError('unexpected response format')
        new_auth = AuthToken.from_login_data(result)
        new_auth.set_cookies(client.cookie_jar)
        return new_auth
    except Exception as e:
        logger.exception(
            f'[login] failed to log in: error when parsing response: {type(e)}, {e}. Raw response: {result}')
        return None


async def fetch_stream_url(client: HttpClient,
                           logger: logging.Logger,
                           stream_id: str,
                           auth: AuthToken) -> Optional[str]:
    url = f'https://www.withny.fun/api/streams/{stream_id}/playback-url'
    headers = {'Referer': url, 'Authorization': auth.plain_token}
    result = await client.request_json(url, headers=headers, settings=RetrySettings(retry_times=2))
    if result is None:
        return None
    if not isinstance(result, str):
        logger.warning(f'unexpected stream url format: {result}')
        return None
    return result


def find_stream_data(data: Optional[JSONType], stream_id: str) -> Optional[dict]:
    if data is None:
        return None
    if not isinstance(data, list):
        return None
    for item in data:
        if not isinstance(item, dict):
            continue
        item_id = item.get('uuid')
        if stream_id != item_id:
            continue
        return item
    return None


Plugins.register('withny.live', Plugins.kind.ASSOCIATED_RECORD)(WithnyRecord)


@Plugins.register('withny.live', Plugins.kind.ACTOR_CONFIG)
class WithnyLiveConfig(TaskActionConfig, BaseDbConfig):
    pass


@Plugins.register('withny.live', Plugins.kind.ACTOR_ENTITY)
class WithnyLiveEntity(TaskActionEntity):
    cookies_file: FilePath
    """path to a text file containing cookies in Netscape format"""
    poll_interval: PositiveInt = 60
    """how often live status of the stream that should have started by now is updated, in seconds"""
    poll_attempts: PositiveInt = 30
    """how many times live status of the stream that should have started by now is updated before giving up"""

    @field_validator('cookies_file')
    @classmethod
    def check_cookies(cls, path: FilePath):
        try:
            load_cookies(path, raise_on_error=True)
        except Exception as e:
            raise ValueError(f'{e}') from e
        return path


@Plugins.register('withny.live', Plugins.kind.ACTOR)
class WithnyLive(TaskAction):
    """
    Wait for livestream on Withny

    If incoming record comes from the "withny" monitor and represents an ongoing or upcoming livestream,
    waits for the stream start and tries to fetch direct `playlist_url` of the stream, then emits updated record
    down the chain. Number and frequency of attempts is limited by `poll_attempts` and `poll_interval` settings.

    Requires cookies from a logged in account to work.
    """
    def __init__(self, conf: WithnyLiveConfig, entities: Sequence[WithnyLiveEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.conf: WithnyLiveConfig
        self.state_storage = StateStorage()
        self.db = RecordDB(conf.db_path, logger=self.logger.getChild('db'))

    async def request_json(self, url: str,
                           base_update_interval: float,
                           current_update_interval: float,
                           client: HttpClient, method='GET',
                           headers: Optional[Dict[str, str]] = None,
                           params: Optional[Any] = None,
                           data: Optional[Any] = None,
                           data_json: Optional[Any] = None) -> Tuple[Optional[JSONType], float]:
        '''Helper method to make http request to a json endpoint'''
        state = self.state_storage.get(url, method, params)
        response = await client.request(url, params, data, data_json, headers, method, state)
        update_interval = response.next_update_interval(base_update_interval, current_update_interval, True)
        if response.has_json():
            data = response.json()
        else:
            data = None
        return data, update_interval

    def parse_record(self, data: Optional[JSONType], parser: Callable) -> Optional[WithnyRecord]:
        if data is None or not isinstance(data, dict):
            return None
        try:
            record = parser(data)
            return record
        except Exception as e:
            self.logger.exception(f'failed to parse stream record: {e}')
            self.logger.debug(f'raw record data: {data}')
            return None

    async def fetch_updated_record(self,
                                   client: HttpClient,
                                   record: WithnyRecord,
                                   update_interval: float,
                                   base_update_interval: float) -> Tuple[Optional[WithnyRecord], float]:
        updated_record = None
        if record.schedule_id is not None:
            if record.scheduled is not None and record.scheduled > utcnow():
                self.logger.debug(f'stream {record.stream_id} is scheduled, using schedule_id')
                url = f'https://www.withny.fun/api/schedules/{record.schedule_id}'
                maybe_record_data, update_interval = await self.request_json(url, base_update_interval,
                                                                             update_interval, client)
                updated_record = self.parse_record(maybe_record_data, parse_schedule_record)
        if record.schedule_id is None or updated_record is None:
            self.logger.debug(f'stream {record.stream_id} is not scheduled or schedule update failed, using stream_id')
            url = f'https://www.withny.fun/api/streams/with-rooms?username={record.username}'
            maybe_data, update_interval = await self.request_json(url, base_update_interval, update_interval, client)
            maybe_record_data = find_stream_data(maybe_data, record.stream_id)
            updated_record = self.parse_record(maybe_record_data, parse_live_record)
            if updated_record is not None:
                if updated_record.schedule_id is not None or updated_record.scheduled is not None:
                    self.logger.warning(f'record {updated_record} should be live but is scheduled: id={updated_record.schedule_id}, scheduled={updated_record.scheduled}')

        return updated_record, update_interval

    async def ensure_login(self, client: HttpClient, entity: WithnyLiveEntity) -> Optional[AuthToken]:
        try:
            auth = AuthToken.from_cookies(client.cookie_jar)
        except ValueError:
            self.logger.warning(f'[{entity.name}] failed to login: error loading data from cookies')
            return None
        if has_expired(auth.refresh_token_expiration):
            # no point checking auth token here, since it does not outlive refresh token
            self.logger.warning(f'[{entity.name}] failed to login: cookies has expired')
            return None
        elif not has_expired(auth.token_expiration):
            self.logger.debug(f'[{entity.name}] skipping logging in, valid auth token is already present in cookies')
            return auth
        else:
            expiration = datetime.timedelta(seconds=(timestamp_now_ms() - int(auth.token_expiration)) / 1000)
            self.logger.debug(f'[{entity.name}] token expired {expiration} ago, refreshing')
            new_auth = await refresh_auth(client, self.logger)
            if new_auth is None:
                return None
            self.logger.debug(f'[{entity.name}] storing refreshed cookies to "{entity.cookies_file}"')
            try:
                save_cookies(client.cookie_jar, str(entity.cookies_file))
            except CookieStoreError as e:
                self.logger.warning(f'[{entity.name}] {e}')
            return new_auth

    async def handle_record_task(self, logger: logging.Logger, client: HttpClient, entity: WithnyLiveEntity,
                                 record: Record, info: TaskStatus) -> None:
        if not isinstance(record, WithnyRecord):
            return

        # wait for the stream to go live, return if there is no point waiting anymore
        update_interval = float(entity.poll_interval)
        for attempt in range(entity.poll_attempts):
            await asyncio.sleep(update_interval)
            msg = f'stream {record.stream_id} by {record.username}, attempt {attempt}: fetching live status'
            self.logger.debug(msg)
            updated_record, update_interval = await self.fetch_updated_record(client, record, update_interval,
                                                                              entity.poll_interval)
            if updated_record is None:
                continue

            updated_record.origin = record.origin
            updated_record.chain = record.chain
            record = updated_record

            if record.end is not None:
                self.logger.debug(f'stream {record.stream_id} has ended at {record.end}: {record!r}')
                msg = 'stream has already ended'
                self.on_record(entity, WithnyLiveErrorEvent(text=msg, record=record))
                return
            elif record.start is not None:
                self.logger.debug(f'stream {record.stream_id}, attempt {attempt}: {record.username} is live')
                break
            elif record.scheduled is None:
                self.logger.debug(
                    f'stream {record.stream_id} has neither start nor scheduled date, dropping record {record!r}')
                msg = 'stream has neither start nor scheduled date'
                self.on_record(entity, WithnyLiveErrorEvent(text=msg, record=record))
                return
            else:
                self.logger.debug(f'stream {record.stream_id}, attempt {attempt}: scheduled {record.scheduled}')
                time_left = record.scheduled - datetime.datetime.now(datetime.timezone.utc)
                delay = time_left.total_seconds()
                if delay > 0:
                    msg = f'stream {record.stream_id} is scheduled to {record.scheduled}, waiting for {time_left}: {record!r}'
                    self.logger.debug(msg)
                    msg = f'waiting for stream to start in {time_left}'
                    info.set_status(msg)
                else:
                    delay = entity.poll_interval
                    msg = f'[{attempt}/{entity.poll_attempts}] waiting for stream to start, next check in {delay:.0f}'
                    info.set_status(msg, record)
                update_interval = delay

        else:
            self.logger.debug(
                f'stream {record.stream_id} is not live after {entity.poll_attempts} checks, dropping record {record!r}')
            msg = f"stream didn't go live after {entity.poll_interval * entity.poll_attempts} seconds"
            self.on_record(entity, WithnyLiveErrorEvent(text=msg, record=record))
            return

        # stream is now live for sure, try fetching stream_url
        auth = await self.ensure_login(client, entity)
        if not auth:
            self.logger.warning(f'login failed, aborting processing')
            self.on_record(entity, WithnyLiveErrorEvent(text='login failed', record=record))
            return
        stream_url = await fetch_stream_url(client, self.logger, record.stream_id, auth)
        if stream_url is None:
            self.logger.warning(f'failed to fetch stream_url for stream {record.stream_id}')
            self.on_record(entity, WithnyLiveErrorEvent(text='retrieving playlist url failed', record=record))
            return

        record.playlist_url = stream_url
        self.db.store_records([record], entity.name)
        self.on_record(entity, record)
